# -*- coding: utf-8 -*-

#
# Collection of functions related to Seqnature offset files
#

import os
import glob

from .g2g_utils import get_logger

LOG = get_logger()


def offset_parse_chromosomes(from_file, to_file):
    """
    Parse the chromosome information files and return a dictionary of chromosomes that contain information about the
    chromosome file offsets.

    It assumed that the chromosome file offsets are names "*_offsets_chr{chr}.txt"

    File format is one line per chromosome, with 2 fields: chromosome chromosome_length

    :param from_file: the chromosome file
    :param to_file: the chromosome file
    :return: a dictionary of chromosomes
    """
    chromosomes = {}

    try:
        if not os.path.exists(from_file):
            raise IOError(from_file)

        fd = open(from_file, 'r')
        for line in fd:
            elem = line.strip().split()
            chromosome = elem[0]
            chromosome_length = int(elem[1])
            chromosome_file_path = os.path.dirname(os.path.abspath(from_file))

            files = glob.glob1(chromosome_file_path, "*_offsets_chr{0}.txt".format(chromosome))
            chromosome_file_path = os.path.join(chromosome_file_path, files[0]) if len(files) == 1 else None
            chromosomes[chromosome] = {'from_chr':chromosome, 'from_length':chromosome_length, 'file_path':chromosome_file_path}
        fd.close()

        if not os.path.exists(to_file):
            raise IOError(to_file)

        fd = open(to_file, 'r')
        for line in fd:
            elem = line.strip().split()
            chromosome = elem[0]

            if chromosome in chromosomes:
                chromosome_length = int(elem[1])
                chromosomes[chromosome]['to_chr'] = chromosome
                chromosomes[chromosome]['to_length'] = chromosome_length
        fd.close()
    except IOError, e:
        message = "Error parsing chromosome files: {0}".format(e.message)
        LOG.info(message)
        return {}

    return chromosomes








# -*- coding: utf-8 -*-

#
# Collection of functions related to VCF files
#
# 1 based


from collections import namedtuple

from .g2g_utils import get_logger
from .g2g_fileutils import open_resource
from .exceptions import G2GVCFError

VCF_FIELDS = ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT', 'SAMPLES']
VCFRecord = namedtuple('VCFRecord', VCF_FIELDS)

GT_DATA_FIELDS = ['ref', 'left', 'right', 'gt', 'fi', 'phase']
GTData = namedtuple('GTData', GT_DATA_FIELDS)

GENOTYPE_UNPHASED = '/'
GENOTYPE_PHASED = '|'

LOG = get_logger()

class VCF(object):
    """
    Simple VCF object for parsing VCF files
    """
    def __init__(self, file_name):
        if not file_name:
            raise G2GVCFError("A filename must be supplied")

        self.file_name = file_name
        self.samples = None
        self.current_line = None
        self.current_record = None
        self.reader = open_resource(file_name)
        self._parse_header()

    def _parse_header(self):
        self.current_line = self.reader.next()

        while self.current_line.startswith('##'):
            self.current_line = self.reader.next()

        if self.current_line.startswith('#'):
            elems = self.current_line.strip().split('\t')
            samples = elems[9:]
            self.samples = dict(zip(samples, (x for x in xrange(len(samples)))))
        else:
            raise G2GVCFError("Improperly formatted VCF file")

    def parse_gt(self, sample):
        if not sample:
            raise G2GVCFError("Sample must contain a value")

        sample_index = self.get_sample_index(sample)
        return parse_gt(self.current_record, sample_index)

    def __iter__(self):
        return self

    def next(self):
        self.current_line = self.reader.next()

        while self.current_line.startswith("#"):
            self.current_line = self.reader.next()

        self.current_record = parse_vcf_line(self.current_line)

        return self.current_record

    def get_sample_index(self, sample):
        if not sample:
            raise G2GVCFError("Sample must contain a value")

        if sample in self.samples:
            return self.samples[sample]

        parse_vcf_line()

        raise G2GVCFError("Unknown sample: '{0}'".format(sample))


def parse_vcf_line(line):
    """
    Parse a line in the VCF file.

    :param line: a line from the VCF file
    :type line: str
    :return: :class:`.vcf.VCFRecord`
    """

    if isinstance(line, basestring):
        if line.startswith('#'):
            return None

        elem = line.strip().split('\t')
    elif isinstance(line, list):
        elem = line

    try:
        quality = int(elem[5])
    except ValueError:
        try:
            quality = float(elem[5])
        except ValueError:
            quality = None

    filter_field = None

    if elem[6] != '.':
        filter_field = elem[6].split(';')

    info = elem[7]

    try:
        fmt = elem[8]
    except IndexError:
        fmt = None
    else:
        if fmt == '.':
            fmt = None

    return VCFRecord(elem[0], int(elem[1]), None if elem[2] == '.' else elem[2], elem[3],
                     elem[4].split(','), quality, filter_field, info, fmt, elem[9:])


def parse_gt(vcf_record, sample_index):
    """
    Parse the GT field within the VCF line.

    :param vcf_record: the VCF record
    :type vcf_record: :class:`.vcf.VCFRecord`
    :param sample_index: the strain or sample index
    :type sample_index: int
    :return: :class:`.vcf.GTData`
    """
    if not sample_index:
        raise G2GVCFError("Sample index must contain a value")

    sample_data = vcf_record.SAMPLES[sample_index]
    gt = None
    fi = None
    left = None
    right = None
    phase = None

    if sample_data != '.':
        gt_index = vcf_record.FORMAT.split(':').index('GT')
        fi_index = vcf_record.FORMAT.split(':').index('FI')

        try:
            # parse the GT field
            gt = sample_data.split(':')[gt_index]

            # make sure a call can be made
            if gt != '.' and gt != './.' and gt != '.|.':
                if GENOTYPE_PHASED in gt:
                    genotypes = map(int, gt.split(GENOTYPE_PHASED))
                    phase = GENOTYPE_PHASED
                elif GENOTYPE_UNPHASED in gt:
                    genotypes = map(int, gt.split(GENOTYPE_UNPHASED))
                    phase = GENOTYPE_UNPHASED
                else:
                    raise ValueError("Unknown phase in GT, {0}".format(gt))

                # assuming no triploids for now
                if genotypes[0] == 0:
                    left = vcf_record.REF
                else:
                    left = vcf_record.ALT[genotypes[0]-1]

                if genotypes[1] == 0:
                    right = vcf_record.REF
                else:
                    right = vcf_record.ALT[genotypes[1]-1]

        except ValueError, ve:
            LOG.debug(ve)
        except IndexError, ie:
            LOG.debug(ie)
        try:
            fi = sample_data.split(':')[fi_index]
        except ValueError, ve:
            LOG.debug(ve)
        except IndexError, ie:
            LOG.debug(ie)

    return GTData(vcf_record.REF, left, right, gt, fi, phase)


def parse_gt_new(vcf_tuple, sample_index):
    """
    Parse the GT field within the VCF line.

    :param vcf_record: the VCF record
    :type vcf_record: :class:`.vcf.VCFRecord`
    :param sample_index: the strain or sample index
    :type sample_index: int
    :return: :class:`.vcf.GTData`
    """
    if not sample_index:
        raise G2GVCFError("Sample index must contain a value")

    sample_data = vcf_tuple[sample_index]
    gt = None
    fi = None
    left = None
    right = None
    phase = None

    if sample_data != '.':
        gt_index = vcf_tuple.format.split(':').index('GT')
        fi_index = vcf_tuple.format.split(':').index('FI')

        try:
            # parse the GT field
            gt = sample_data.split(':')[gt_index]

            # make sure a call can be made
            if gt != '.' and gt != './.' and gt != '.|.':
                if GENOTYPE_PHASED in gt:
                    genotypes = map(int, gt.split(GENOTYPE_PHASED))
                    phase = GENOTYPE_PHASED
                elif GENOTYPE_UNPHASED in gt:
                    genotypes = map(int, gt.split(GENOTYPE_UNPHASED))
                    phase = GENOTYPE_UNPHASED
                else:
                    raise ValueError("Unknown phase in GT, {0}".format(gt))

                # assuming no triploids for now
                if genotypes[0] == 0:
                    left = vcf_tuple.ref
                else:
                    left = vcf_tuple.alt.split(',')[genotypes[0]-1]

                if genotypes[1] == 0:
                    right = vcf_tuple.ref
                else:
                    right = vcf_tuple.alt.split(',')[genotypes[1]-1]

        except ValueError, ve:
            LOG.debug(ve)
        except IndexError, ie:
            LOG.debug(ie)
        try:
            fi = sample_data.split(':')[fi_index]
        except ValueError, ve:
            LOG.debug(ve)
        except IndexError, ie:
            LOG.debug(ie)

    return GTData(vcf_tuple.ref, left, right, gt, fi, phase)



# -*- coding: utf-8 -*-

#
# Collection of functions related to BAM and SAM files
#

# pysam uses 0-based coordinates

from collections import namedtuple
import os
import re
import sys

import pysam

from .chain import ChainFile
from .exceptions import G2GCigarFormatError, G2GBAMError, G2GValueError
from .g2g_utils import get_logger
import g2g_fileutils as g2g_fu


FLAG_NONE = 0x0             # base value
FLAG_PAIRED = 0x1           # template having multiple segments in sequencing
FLAG_PROPER_PAIR = 0x2      # each segment properly aligned according to the aligner
FLAG_UNMAP = 0x4            # segment unmapped
FLAG_MUNMAP = 0x8           # next segment in the template unmapped (mate unmapped)
FLAG_REVERSE = 0x10         # SEQ being reverse complemented
FLAG_MREVERSE = 0x20        # SEQ of the next segment in the template being reversed
FLAG_READ1 = 0x40           # the first segment in the template
FLAG_READ2 = 0x80           # the last segment in the template
FLAG_SECONDARY = 0x100      # secondary alignment
FLAG_QCFAIL = 0x200         # not passing quality controls
FLAG_DUP = 0x400            # PCR or optical duplicate
FLAG_SUPPLEMENTARY = 0x800  # supplementary alignment

REGEX_CIGAR = re.compile("(\d+)([\w=])")
REGEX_CIGAR_LENGTH = re.compile("\D")

CIGAR_M = 'M'
CIGAR_I = 'I'
CIGAR_D = 'D'
CIGAR_N = 'N'
CIGAR_S = 'S'
CIGAR_H = 'H'
CIGAR_P = 'P'
CIGAR_E = '='
CIGAR_X = 'X'

CIGAR_m = 0
CIGAR_i = 1
CIGAR_d = 2
CIGAR_n = 3
CIGAR_s = 4
CIGAR_h = 5
CIGAR_p = 6
CIGAR_e = 7
CIGAR_x = 8

CIGAR_N2C = {
    0: 'M',  # alignment match (can be a sequence match or mismatch)
    1: 'I',  # insertion to the reference
    2: 'D',  # deletion from the reference
    3: 'N',  # skipped region from the reference
    4: 'S',  # soft clipping (clipped sequences present in SEQ)
    5: 'H',  # hard clipping (clipped sequences NOT present in SEQ)
    6: 'P',  # padding (silent deletion from padded reference)
    7: '=',  # sequence match
    8: 'X',  # sequence mismatch
    '0': 'M',
    '1': 'I',
    '2': 'D',
    '3': 'N',
    '4': 'S',
    '5': 'H',
    '6': 'P',
    '7': '=',
    '8': 'X'
}

CIGAR_C2N = {
    'M': 0,
    'I': 1,
    'D': 2,
    'N': 3,
    'S': 4,
    'H': 5,
    'P': 6,
    '=': 7,
    'X': 8
}

LOG = get_logger()

Cigar = namedtuple('Cigar', ['code', 'length', 'start', 'end'])


def convert_bam_file(chain_file, file_in, file_out, reverse=False):
    """
    Convert genome coordinates (in BAM/SAM format) between assemblies.  These coordinates
    are stored in the :class:`.chain.ChainFile` object.

    :param chain_file: chain file used for conversion
    :type chain_file: :class:`.chain.ChainFile`
    :param str file_in: the input SAM or BAM file
    :type file_in: string
    :param file_out: the output SAM or file
    :type file_out: string
    :param reverse: reverse direction of original chain file
    :type reverse: boolean
    """
    if not isinstance(chain_file, ChainFile):
        chain_file = g2g_fu.check_file(chain_file)

    if not isinstance(file_in, pysam.Samfile):
        file_in = g2g_fu.check_file(file_in)

    output_file_name = g2g_fu.check_file(file_out, 'w')
    unmapped_file_name = "{0}.unmapped".format(output_file_name)

    LOG.info("CHAIN FILE: {0}".format(chain_file))
    LOG.info("INPUT FILE: {0}".format(file_in))
    LOG.info("OUTPUT FILE: {0}".format(output_file_name))
    LOG.info("UNMAPPED FILE: {0}".format(unmapped_file_name))

    if not isinstance(chain_file, ChainFile):
        LOG.info("Parsing chain file...")
        chain_file = ChainFile(chain_file, reverse=reverse)
        LOG.info("Chain file parsed")

    if not isinstance(file_in, pysam.Samfile):
        try:
            sam_file = pysam.Samfile(file_in, 'rb')
            if len(sam_file.header) == 0:
                raise G2GBAMError("BAM File has no header information")
        except:
            sam_file = pysam.Samfile(file_in, 'r')
            if len(sam_file.header) == 0:
                raise G2GBAMError("SAM File has no header information")

    LOG.info("Converting BAM file")

    new_header = sam_file.header

    # replace 'HD'
    new_header['HD'] = {'VN': 1.0, 'SO': 'coordinate'}

    # replace SQ
    tmp = []
    name_to_id = {}
    id = 0
    for ref_name in sorted(chain_file.chrom_size_to):
        tmp.append({'LN': chain_file.chrom_size_from[ref_name], 'SN': ref_name})
        name_to_id[ref_name] = id
        id += 1

    new_header['SQ'] = tmp

    if 'PG' not in new_header:
        new_header['PG'] = []

    new_header['PG'].append({'ID': 'gtgtools', 'VN': 1.0})

    if 'CO' not in new_header:
        new_header['CO'] = []

    new_header['CO'].append("Original file: {0}".format(file_in))
    new_header['CO'].append("Chain File: {0}".format(chain_file.file_name))

    dir, temp_file_name = os.path.split(file_out)
    parts = temp_file_name.split('.')
    ext = parts[-1]

    if ext.lower() == 'bam':
        new_file = pysam.Samfile(file_out, 'wb', header=new_header)
        new_file_unmapped = pysam.Samfile(unmapped_file_name, 'wb', template=sam_file)
    elif ext.lower() == 'sam':
        new_file = pysam.Samfile(file_out, 'wh', header=new_header)
        new_file_unmapped = pysam.Samfile(unmapped_file_name, 'wh', template=sam_file)
    else:
        raise G2GBAMError("Unable to create new file based upon file extension")

    total = 0
    total_unmapped = 0
    total_fail_qc = 0

    map_statistics = {'total': 0,
                      'fail_cannot_map': 0,
                      'success_simple': 0,
                      'success_complex': 0}

    map_statistics_pair = {'total': 0,
                           'fail_cannot_map': 0,
                           'success_1_fail_2_simple': 0,
                           'success_1_fail_2_complex': 0,
                           'success_1_simple_2_fail': 0,
                           'success_1_simple_2_simple': 0,
                           'success_1_simple_2_complex': 0,
                           'success_1_complex_2_fail': 0,
                           'success_1_complex_2_simple': 0,
                           'success_1_complex_2_complex': 0}

    try:
        while True:
            if total and total % 10000 == 0:
                status_success = 0
                status_failed = 0

                for k, v in map_statistics_pair.iteritems():
                    if k.startswith('success'):
                        status_success += v
                    elif k.startswith('fail'):
                        status_failed += v

                LOG.info("Processed {0:,} reads, {1:,} successful, {2:,} failed".format(total, status_success, status_failed))

            alignment = sam_file.next()
            alignment_new = pysam.AlignedRead()
            read_chr = sam_file.getrname(alignment.tid)

            # READ ONLY

            # aend                  aligned reference position of the read on the reference genome
            # alen                  aligned length of the read on the reference genome.
            # positions             a list of reference positions that this read aligns to
            # qend                  end index of the aligned query portion of the sequence (0-based, exclusive)
            # qlen                  Length of the aligned query sequence
            # qqual                 aligned query sequence quality values
            # qstart                start index of the aligned query portion of the sequence (0-based, inclusive)
            # query                 aligned portion of the read and excludes any flanking bases that were soft clipped
            # rlen                  length of the read

            # TRUE / FALSE (setting effects flag)

            # is_paired             true if read is paired in sequencing
            # is_proper_pair        true if read is mapped in a proper pair
            # is_qcfail             true if QC failure
            # is_read1              true if this is read1
            # is_read2              true if this is read2
            # is_reverse            true if read is mapped to reverse strand
            # is_secondary          true if not primary alignment
            # is_unmapped           true if read itself is unmapped
            # mate_is_reverse       true is read is mapped to reverse strand
            # mate_is_unmapped      true if the mate is unmapped

            # SET

            # cigar                 cigar as list of tuples
            # cigarstring           alignment as a string
            # flag                  properties flag
            # mapq                  mapping quality
            # pnext                 the position of the mate
            # pos                   0-based leftmost coordinate
            # pnext                 the position of the mate
            # qname                 the query name
            # rnext                 the reference id of the mate
            # seq                   read sequence bases, including soft clipped bases
            # tid                   target id, contains the index of the reference sequence in the sequence dictionary

            # DON'T NEED TO SET or SHOULD WE SET?

            # qual                  read sequence base qualities, including soft clipped bases
            # tags                  the tags in the AUX field
            # tlen                  insert size

            total += 1

            LOG.debug('~'*80)
            LOG.debug("Converting {0} {1} {2} {3}".format(alignment.qname, read_chr, alignment.pos, alignment.cigarstring))

            if alignment.is_qcfail:
                LOG.debug("\tFail due to qc of old alignment")
                new_file_unmapped.write(alignment)
                total_fail_qc += 1
                continue

            if alignment.is_unmapped:
                LOG.debug("\tFail due to unmapped old alignment")
                new_file_unmapped.write(alignment)
                total_unmapped += 1
                continue

            if not alignment.is_paired:
                LOG.debug("SINGLE END ALIGNMENT")
                map_statistics['total'] += 1

                alignment_new.seq = alignment.seq
                alignment_new.flag = FLAG_NONE
                alignment_new.mapq = alignment.mapq
                alignment_new.qname = alignment.qname
                alignment_new.qual = alignment.qual
                alignment_new.tags = alignment.tags

                read_start = alignment.pos
                read_end = alignment.aend
                read_strand = '-' if alignment.is_reverse else '+'

                mappings = chain_file.find_mappings(read_chr, read_start, read_end)

                # unmapped
                if mappings is None:
                    LOG.debug("\tFail due to no mappings")
                    new_file_unmapped.write(alignment)
                    map_statistics['fail_cannot_map'] += 1

                elif len(mappings) == 1:
                    if alignment.is_reverse:
                        alignment_new.flag |= FLAG_REVERSE

                    alignment_new.tid = name_to_id[mappings[0].to_chr]
                    alignment_new.pos = mappings[0].to_start
                    alignment_new.cigar = alignment.cigar
                    new_file.write(alignment_new)

                    LOG.debug("\tSuccess (simple): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    map_statistics['success_simple'] += 1

                else:
                    LOG.debug("MAPPINGS: {0}".format(len(mappings)))
                    for m in mappings:
                        LOG.debug("> {0}".format(m))

                    if alignment.is_reverse:
                        alignment_new.flag |= FLAG_REVERSE

                    alignment_new.tid = name_to_id[mappings[0].to_chr]
                    alignment_new.pos = mappings[0].to_start
                    alignment_new.cigar = convert_cigar(alignment.cigar, read_chr, chain_file, alignment.seq, read_strand, alignment.pos)
                    new_file.write(alignment_new)

                    LOG.debug("\tSuccess (complex): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    map_statistics['success_complex'] += 1

            else:
                LOG.debug("PAIRED END ALIGNMENT")
                map_statistics_pair['total'] += 1

                alignment_new.seq = alignment.seq
                alignment_new.flag = FLAG_PAIRED
                alignment_new.mapq = alignment.mapq
                alignment_new.qname = alignment.qname
                alignment_new.qual = alignment.qual
                alignment_new.tags = alignment.tags

                if alignment.is_read1:
                    alignment_new.flag |= FLAG_READ1
                if alignment.is_read2:
                    alignment_new.flag |= FLAG_READ2

                if alignment.is_reverse:
                    alignment_new.flag |= FLAG_REVERSE
                if alignment.mate_is_reverse:
                    alignment_new.flag |= FLAG_MREVERSE

                read1_chr = sam_file.getrname(alignment.tid)
                read1_start = alignment.pos
                read1_end = alignment.aend
                read1_strand = '-' if alignment.is_reverse else '+'
                read1_mappings = chain_file.find_mappings(read1_chr, read1_start, read1_end) #, read1_strand)

                read2_chr = None
                read2_start = None
                read2_end = None
                read2_strand = None
                read2_mappings = None

                if alignment.mate_is_unmapped:
                    alignment_new.flag |= FLAG_MUNMAP
                else:
                    read2_chr = sam_file.getrname(alignment.rnext)
                    read2_start = alignment.pnext
                    read2_end = read2_start + 1
                    read2_strand = '-' if alignment.mate_is_reverse else '+'
                    try:
                        read2_mappings = chain_file.find_mappings(read2_chr, read2_start, read2_end)
                    except:
                        read2_mappings = None

                if read1_mappings is None and read2_mappings is None:

                    alignment_new.flag |= FLAG_UNMAP
                    alignment_new.flag |= FLAG_MUNMAP

                    LOG.debug("\tFail due to no mappings")
                    new_file_unmapped.write(alignment)
                    map_statistics_pair['fail_cannot_map'] += 1

                elif read1_mappings is None and read2_mappings and len(read2_mappings) == 1:

                    alignment_new.flag |= FLAG_UNMAP

                    alignment_new.pos = 0
                    alignment_new.cigarstring = '0M'
                    alignment_new.rnext = name_to_id[read2_mappings[0].to_chr]
                    alignment_new.pnext = read2_mappings[0].to_start
                    alignment_new.tlen = 0

                    LOG.debug("\tPair Success (1:fail,2:simple): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    new_file.write(alignment_new)
                    map_statistics_pair['success_1_fail_2_simple'] += 1

                elif read1_mappings is None and read2_mappings and len(read2_mappings) > 1:

                    alignment_new.flag |= FLAG_UNMAP

                    alignment_new.pos = 0
                    alignment_new.cigarstring = '0M'
                    alignment_new.rnext = name_to_id[read2_mappings[0].to_chr]
                    alignment_new.pnext = read2_mappings[0].to_start
                    alignment_new.tlen = 0

                    LOG.debug("\tPair Success (1:fail,2:complex): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    new_file.write(alignment_new)
                    map_statistics_pair['success_1_fail_2_complex'] += 1

                elif read1_mappings and len(read1_mappings) == 1 and read2_mappings is None:

                    alignment_new.flag |= FLAG_MUNMAP

                    alignment_new.tid = name_to_id[read1_mappings[0].to_chr]
                    alignment_new.pos = read1_mappings[0].to_start
                    alignment_new.cigar = alignment.cigar

                    alignment_new.rnext = name_to_id[read1_mappings[0].to_chr]
                    alignment_new.pnext = 0
                    alignment_new.tlen = 0    # CHECK

                    LOG.debug("\tPair Success (1:simple,2:fail): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    new_file.write(alignment_new)
                    map_statistics_pair['success_1_simple_2_fail'] += 1

                elif read1_mappings and len(read1_mappings) == 1 and read2_mappings and len(read2_mappings) == 1:

                    alignment_new.tid = name_to_id[read1_mappings[0].to_chr]
                    alignment_new.pos = read1_mappings[0].to_start
                    alignment_new.cigar = alignment.cigar

                    alignment_new.rnext = name_to_id[read2_mappings[0].to_chr]
                    alignment_new.pnext = read2_mappings[0].to_start
                    alignment_new.tlen = 0    # CHECK

                    LOG.debug("\tPair Success (1:simple,2:simple): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    new_file.write(alignment_new)
                    map_statistics_pair['success_1_simple_2_simple'] += 1

                elif read1_mappings and len(read1_mappings) == 1 and read2_mappings and len(read2_mappings) > 1:

                    alignment_new.tid = name_to_id[read1_mappings[0].to_chr]
                    alignment_new.pos = read1_mappings[0].to_start
                    alignment_new.cigar = alignment.cigar

                    alignment_new.rnext = name_to_id[read2_mappings[0].to_chr]
                    alignment_new.pnext = read2_mappings[0].to_start
                    alignment_new.tlen = 0    # CHECK

                    LOG.debug("\tPair Success (1:simple,2:complex): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    new_file.write(alignment_new)
                    map_statistics_pair['success_1_simple_2_complex'] += 1

                elif read1_mappings and len(read1_mappings) > 1 and read2_mappings is None:

                    alignment_new.flag |= FLAG_MUNMAP

                    alignment_new.tid = name_to_id[read1_mappings[0].to_chr]
                    alignment_new.pos = read1_mappings[0].to_start
                    alignment_new.cigar = convert_cigar(alignment.cigar, read_chr, chain_file, alignment.seq, read1_strand, alignment.pos)

                    alignment_new.rnext = name_to_id[read1_mappings[0].to_chr]
                    alignment_new.pnext = 0
                    alignment_new.tlen = 0    # CHECK

                    LOG.debug("\tPair Success (1:complex,2:fail): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    new_file.write(alignment_new)
                    map_statistics_pair['success_1_complex_2_fail'] += 1

                elif read1_mappings and len(read1_mappings) > 1 and read2_mappings and len(read2_mappings) == 1:

                    alignment_new.tid = name_to_id[read1_mappings[0].to_chr]
                    alignment_new.pos = read1_mappings[0].to_start
                    alignment_new.cigar = convert_cigar(alignment.cigar, read_chr, chain_file, alignment.seq, read1_strand, alignment.pos)

                    alignment_new.rnext = name_to_id[read2_mappings[0].to_chr]
                    alignment_new.pnext = read2_mappings[0].to_start
                    alignment_new.tlen = 0    # CHECK

                    LOG.debug("\tPair Success (1:complex,2:simple): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    new_file.write(alignment_new)
                    map_statistics_pair['success_1_complex_2_simple'] += 1

                elif read1_mappings and len(read1_mappings) > 1 and read2_mappings and len(read2_mappings) > 1:

                    alignment_new.tid = name_to_id[read1_mappings[0].to_chr]
                    alignment_new.pos = read1_mappings[0].to_start
                    alignment_new.cigar = convert_cigar(alignment.cigar, read_chr, chain_file, alignment.seq, read1_strand, alignment.pos)

                    alignment_new.rnext = name_to_id[read2_mappings[0].to_chr]
                    alignment_new.pnext = read2_mappings[0].to_start
                    alignment_new.tlen = 0    # CHECK

                    LOG.debug("\tPair Success (1:complex,2:complex): {0} {1}".format(alignment_new.pos, alignment_new.cigarstring))
                    new_file.write(alignment_new)
                    map_statistics_pair['success_1_complex_2_complex'] += 1

                else:
                    raise G2GBAMError("Unknown BAM/SAM conversion/parse situation")

    except StopIteration:
        LOG.info("All reads processed")

    LOG.info("  {:>10} TOTAL ENTRIES".format(total))
    LOG.info("  {:>10} TOTAL UNMAPPED ".format(total_unmapped))
    LOG.info("  {:>10} TOTAL FAIL QC ".format(total_fail_qc))

    if map_statistics['total'] > 0:
        LOG.info("")
        LOG.info("Mapping Summary Single End")
        LOG.info("  {:>10} TOTAL ENTRIES".format(map_statistics['total']))
        LOG.info("")
        LOG.info("  {:>10} TOTAL SUCCESS".format(map_statistics['success_simple'] + map_statistics['success_complex']))
        LOG.info("  {:>10} Simple".format(map_statistics['success_simple']))
        LOG.info("  {:>10} Complex".format(map_statistics['success_complex']))
        LOG.info("")
        LOG.info("  {:>10} TOTAL FAILURES".format(map_statistics['fail_cannot_map']))
        LOG.info("  {:>10} Cannot Map ".format(map_statistics['fail_cannot_map']))

    if map_statistics_pair['total'] > 0:
        total_success = 0
        for k, v in map_statistics_pair.iteritems():
            if k.startswith('success'):
                total_success += v

        LOG.info("")
        LOG.info("Mapping Summary Paired End")
        LOG.info("  {:>10} TOTAL ENTRIES".format(map_statistics_pair['total']))
        LOG.info("")
        LOG.info("  {:>10} TOTAL SUCCESS".format(total_success))
        LOG.info("  {:>10} Read 1 Failed, Read 2 Simple".format(map_statistics_pair['success_1_fail_2_simple']))
        LOG.info("  {:>10} Read 1 Failed, Read 2 Complex".format(map_statistics_pair['success_1_fail_2_complex']))
        LOG.info("  {:>10} Read 1 Simple, Read 2 Failed".format(map_statistics_pair['success_1_simple_2_fail']))
        LOG.info("  {:>10} Read 1 Simple, Read 2 Simple".format(map_statistics_pair['success_1_simple_2_simple']))
        LOG.info("  {:>10} Read 1 Simple, Read 2 Complex".format(map_statistics_pair['success_1_simple_2_complex']))
        LOG.info("  {:>10} Read 1 Complex, Read 2 Failed".format(map_statistics_pair['success_1_complex_2_fail']))
        LOG.info("  {:>10} Read 1 Complex, Read 2 Simple".format(map_statistics_pair['success_1_complex_2_simple']))
        LOG.info("  {:>10} Read 1 Complex, Read 2 Complex".format(map_statistics_pair['success_1_complex_2_complex']))
        LOG.info("")
        LOG.info("  {:>10} TOTAL FAILURES".format(map_statistics_pair['fail_cannot_map']))
        LOG.info("  {:>10} Cannot Map".format(map_statistics_pair['fail_cannot_map']))
        LOG.info("")

    LOG.info("BAM File Converted")


#
# Functions dealing with CIGAR strings
#
#
#    BAM	OP	    Description
#    0	    M	    alignment match
#    1	    I	    insertion to reference
#    2	    D	    deletion from reference. region deleted from reference genome
#    3	    N	    skipped region from the reference
#    4	    S	    soft clipping (clipped sequence present in SEQ)
#    5	    H	    hard clipping (clipped sequences NOT present in SEQ)
#    6	    P	    padding (silent deletion from padded reference)
#    7	    =	    sequence match
#    8	    X	    sequence mismatch
#

def cigarlist_to_cigarstring(cigar_list):
    """
    Convert a list of tuples into a cigar string.

    Example::

             [ (0, 10), (1, 1), (0, 75), (2, 2), (0, 20) ]
        =>       10M      1I      75M      2D      20M
        =>                10M1I75M2D20M


    :param cigar_list: a list of tuples (code, length)
    :type cigar_list: list
    :return: the cigar string
    :rtype: string
    :raises: :class:`.exceptions.G2GCigarFormatError` on invalid cigar string
    """
    cigar = ''
    if isinstance(cigar_list, Cigar):
        try:
            for i in cigar_list:
                cigar += str(i.length) + i.code
        except KeyError:
            raise G2GCigarFormatError("Invalid cigar code: " + str(i))
    else:
        try:
            for i in cigar_list:
                cigar += str(i[1]) + CIGAR_N2C[i[0]]
        except KeyError:
            raise G2GCigarFormatError("Invalid cigar code: " + str(i))

    return cigar

def cigar_to_string(cigar):
    """
    Convert a list of tuples into a cigar string.

    Example::

             [ (0, 10), (1, 1), (0, 75), (2, 2), (0, 20) ]
        =>       10M      1I      75M      2D      20M
        =>                10M1I75M2D20M


    :param cigar_list: a list of tuples (code, length)
    :type cigar_list: list
    :return: the cigar string
    :rtype: string
    :raises: :class:`.exceptions.G2GCigarFormatError` on invalid cigar string
    """
    cigar = ''
    try:
        for i in cigar:
            cigar += str(i.length) + i.code
    except KeyError:
        raise G2GCigarFormatError("Invalid cigar code: " + str(i))

    return cigar

def _cigar_to_list(cigar_string):
    """
    Convert a list of tuples into a cigar string

    Example::

                         10M1I75M2D20M
        =>       10M      1I      75M      2D      20M
        =>   [ (0, 10), (1, 1), (0, 75), (2, 2), (0, 20) ]


    :param cigar_string: a cigar string
    :return: a list of tuples (code, length)
    :rtype: list
    :raises: :class:`.exceptions.G2GCigarFormatError` on invalid cigar string
    """
    matches = REGEX_CIGAR.findall(cigar_string)
    possible_length = len(REGEX_CIGAR_LENGTH.findall(cigar_string))

    if len(matches) != possible_length:
        raise G2GCigarFormatError("Invalid cigar string: {0}".format(cigar_string))

    lst = []
    try:
        for m in matches:
            lst.append(1)#(CIGAR_CODES_REV[m[1]], int(m[0])))
    except KeyError:
        raise G2GCigarFormatError("Invalid cigar string: {0} : {1} ".format(cigar_string, str(m)))

    return lst


def _cigar_convert(cigar, chromosome, chain, strand='+', position=0):
    """
    PHASE 1

    Convert each CIGAR element to new mappings and construct an array on NEW cigar elements

    For example, depending on the Intervals in the CHAIN file, let's say we have the following
    CIGAR string: 35M49N65M

    This could get converted into
    35M ==> 4M150D31M
    49N ==> -1N (remember, surrounding M's are used to find the length of N which is done on next pass)
    65M ==> 65M

    First pass yields: 35M49N65M => 4M150D31M-1N65M

    :param cigar:
    :param chromosome:
    :param chain:
    :param strand:
    :param position:
    :return:
    """
    cigar_new = []
    current_pos = position
    cigar_no = 0

    for c in cigar:
        cigar_no += 1

        LOG.debug("Element #{0}, '{1}{2}' specified, location: {3}".format(cigar_no, c[1], CIGAR_N2C[c[0]], current_pos))

        increment = c[1]

        if c[0] == CIGAR_m:
            new_mappings = chain.find_mappings(chromosome, current_pos, current_pos + c[1])

            if not new_mappings:
                LOG.debug("Mappings: None")
                cigar_new.append(Cigar(CIGAR_S, c[1], 0, 0))
            elif len(new_mappings) == 1:
                LOG.debug("Mappings: Easy: {0}".format(new_mappings[0]))
                cigar_new.append(Cigar(CIGAR_M, new_mappings[0].to_end - new_mappings[0].to_start, new_mappings[0].to_start, new_mappings[0].to_end))
            else:
                # multiple maps, not so easy
                last = None

                for m in new_mappings:
                    LOG.debug("Mappings: Multiple: {0}".format(m))
                    if not last:
                        last = m

                        if current_pos < m.from_start:
                            # special case of first match not in interval, handle accordingly
                            LOG.debug("Adding 'S', because {0} < {1}".format(current_pos, m.from_start))
                            cigar_new.append(Cigar(CIGAR_S, m.from_start - current_pos, 0, 0))
                    else:
                        if m.from_start != last.from_end:
                            LOG.debug("Adding 'M' and 'I', because {0} != {1}".format(m.from_start, last.from_end))

                            cigar_new.append(Cigar(CIGAR_M, last.to_end - last.to_start, last.to_start, last.to_end))
                            cigar_new.append(Cigar(CIGAR_I, m.from_start - last.from_end, last.to_start, last.to_end))

                        elif m.to_start != last.to_end:
                            LOG.debug("Adding 'M' and 'D', because {0} != {1}".format(m.to_start, last.to_end))

                            cigar_new.append(Cigar(CIGAR_M, last.to_end - last.to_start, last.to_start, last.to_end))
                            cigar_new.append(Cigar(CIGAR_D, m.to_start - last.to_end, 0, 0))

                        last = m

                LOG.debug("Adding 'M'")
                cigar_new.append(Cigar(CIGAR_M, last.to_end - last.to_start, last.to_start, last.to_end))

        elif c[0] == CIGAR_i:

            LOG.debug("Adding 'I' and 'D'")

            cigar_new.append(Cigar(CIGAR_I, c[1], 0, 0))
            cigar_new.append(Cigar(CIGAR_D, -1, 0, 0))

            increment = 0

        elif c[0] == CIGAR_d:

            LOG.debug("Adding 'D'")

            cigar_new.append(Cigar(CIGAR_D, -1, 0, 0))

        elif c[0] == CIGAR_n:

            LOG.debug("Adding 'N'")
            cigar_new.append(Cigar(CIGAR_N, -1, 0, 0))

        elif c[0] in [CIGAR_s, CIGAR_h]:

            LOG.debug("Adding '{0}'".format(CIGAR_N2C[c[0]]))
            cigar_new.append(Cigar(CIGAR_N2C[c[0]], c[1], 0, 0))

        else:
            # other
            LOG.debug("OTHER CODE '{0}' found, looking at {1} at {2}".format(CIGAR_N2C[c[0]], c, current_pos))
            raise G2GCigarFormatError("ERROR: Not handling the values in this cigar string: {0}".format(cigar))

        #current_pos += c[1]
        current_pos += increment

        LOG.debug("Current CIGAR: {0}".format(cigar_new))

    return cigar_new


def _cigar_combine_consecutive(cigar):
    """
    Combine consecutive features in a cigar string.

    For example, 2 N's become 1

    :param cigar:
    :return:
    """
    done = False

    while not done:
        done = True

        for i in xrange(0, len(cigar)-1):

            LOG.debug("{0}={1}".format(i, cigar[i]))
            LOG.debug("{0}={1}".format(i+1, cigar[i+1]))
            if cigar[i].code == cigar[i+1].code:
                done = False
                break

        if not done:
            cigar_temp = []
            cigar_temp.extend(cigar[:i])

            cm1 = cigar[i]
            cm2 = cigar[i+1]
            cm_new = Cigar(cm1.code, cm1.length + cm2.length, cm1.start, cm2.end)
            cigar_temp.append(cm_new)

            LOG.debug("Found consecutive elements {0} and {1}, combined into {2}".format(cm1, cm2, cm_new))

            cigar_temp.extend(cigar[i+2:])

            cigar = cigar_temp

    return cigar


def _cigar_fix_pre_and_post_M(cigar):
    """

    :param cigar:
    :return:
    """
    # pre M to S fix
    for i in xrange(0, len(cigar)):
        if cigar[i].code == CIGAR_M:
            break

    if i != 0:
        first_m = i

        length = 0
        for i in xrange(0, first_m):
            if cigar[i].code in [CIGAR_I, CIGAR_S, CIGAR_H]:
                length += cigar[i].length

        temp_cigar = [Cigar(CIGAR_S, length, 0, 0)]
        temp_cigar.extend(cigar[i+1:])

        cigar = temp_cigar

    # post M to S fix
    for i in reversed(xrange(0, len(cigar))):
        if cigar[i].code == CIGAR_M:
            break

    if i > 0 and i != (len(cigar) - 1):
        last_m = i

        length = 0
        for i in xrange(last_m+1, len(cigar)):
            if cigar[i].code in [CIGAR_M, CIGAR_I, CIGAR_S]:
                length += cigar[i].length

        temp_cigar = []
        temp_cigar.extend(cigar[:i-1])
        temp_cigar.append(Cigar(CIGAR_S, length, 0, 0))
        cigar = temp_cigar

    return cigar


def _cigar_remove_softs_between_m(cigar):
    """
    Remove soft if surrounded by Ms

    :param cigar:
    :return:
    """
    done = False

    while not done:
        done = True

        for i in xrange(1, len(cigar)-1):
            if cigar[i].code == CIGAR_S:
                done = False
                break

        if done:
            break

        before = None
        after = None
        for x in reversed(xrange(i)):
            if cigar[x].code == CIGAR_M:
                before = cigar[x]
                break

        for x in xrange(i+1, len(cigar)):
            if cigar[x].code == CIGAR_M:
                after = cigar[x]
                break

        if before and after:
            LOG.debug("Found 'S' between 'M' so removing 'S'")
            cigar_temp = []
            cigar_temp.extend(cigar[:i])
            cigar_temp.extend(cigar[i+1:])
            cigar = cigar_temp
            LOG.debug(cigar)
        else:
            done = True

    return cigar


def _cigar_fix_lengths(cigar, sequence):
    """

    :return:
    """
    # Assign length to -1's
    #
    # Since N's aren't mapped we look at the surrounding M's to find the length of the N's
    #
    # Example: 35M49N65M ==> 4M150D31M-1N65M, the -1 will be corrected by finding the last position of the previous
    # M and first position of the next M
    #
    # there are a few special cases that are handled
    # since there were multiple mappings, we will need to figure out the location on the N's
    done = False

    while not done:
        done = True

        # find first element without a length
        i = 0
        for cm in cigar:
            if cm.length == -1:
                break
            i += 1

        if i == len(cigar):
            done = True
            break

        LOG.debug("Found '{0}' at {1}: {2}".format(cm.code, i, cm))

        before = None
        after = None

        # Simple case is surrounded by mapping positions, but might not be the case
        for x in reversed(xrange(i)):
            if cigar[x].code == CIGAR_M:
                before = cigar[x]
                break


        for x in xrange(i+1, len(cigar)):
            if cigar[x].code == CIGAR_M:
                after = cigar[x]
                break

        # special case of 89M2000N11M
        # what happens when thi sis converted to 89M-1N11S (no M at end)
        # we should have 89M11S

        LOG.debug("Before: {0}".format(before))
        LOG.debug("After: {0}".format(after))

        # check if all cigar elements from here to end do not have a length
        a = i
        while a < len(cigar) - 1:
            if cigar[a].length != -1:
                break
            a += 1

        # if a == len(cigar_mapping) -1 than all the rest have no length
        LOG.debug("a={0}, len(cigar_mapping) - 1={1}".format(a, len(cigar) - 1))
        if (a == len(cigar) - 1 and cigar[a].start == -1) or not after or not before:
            # take the rest as a clip
            LOG.debug("Found a clip")
            temp_cigar_mappings = cigar[:i]
            temp_total = 0
            for t in temp_cigar_mappings:
                if t.code in [CIGAR_M, CIGAR_I, CIGAR_S]:
                    temp_total += t.length

            temp_cigar_mappings.append(Cigar(CIGAR_S, len(sequence) - temp_total, -1, -1))
            cigar = temp_cigar_mappings
            done = True
        else:
            c = cigar[i]
            new_c = Cigar(c.code, after.start - before.end, before.end, after.start)
            LOG.debug("Replacing, old = {0}, new = {1}".format(c, new_c))
            cigar[i] = new_c

            done = False

    LOG.debug("Removing 0 length elements, if any")
    new_cigar = []
    for cm in cigar:
        if cm[1] == 0:
            LOG.debug("Removing {}".format(cm))
            continue
        new_cigar.append(cm)

    return new_cigar


def convert_cigar(cigar, chromosome, chain, sequence, strand='+', position=0):
    """
    Generate the cigar string of an old alignment.

    P1: Map M with bx; Inherit S and H; Inherit I but put -1D right behind it; Put -1D or -1N when it’s there.
        P1a: Convert xM, if it has zero length after mapping, to xS

    P2: Remove S (including new S originate from unmapped M) if it is surrounded by any pair of consecutive Ms that
    survived P2

    P3: Adjust the size of D or N that are inbetween Ms. Remove it if they have zero length.

    P4: Combine duplicated entries (I guess mostly M or S)

    P5: Put yS for the unmapped regions before the first M and/or after the last M (I believe adding S, H, I’s in
    those regions should get you y). In this phase remove the remaining -1D or -1N in those regions first.

    :param old_alignment: the old alignment
    :type old_alignment: :class:`pysam.AlignedRead`
    :param chromosome:  the chromosome
    :type chromosome: string
    :param chain: the chain file
    :type chain: the chain file
    :return: a new cigar string based upon the mappings
    :raises: :class:`.exceptions.G2GCigarFormatError` on invalid cigar string
    """

    old_cigar = cigarlist_to_cigarstring(cigar)
    LOG.debug("CIGAR CONVERSION : {0}".format(old_cigar))

    #
    # PHASE 1: Convert each CIGAR element to new mappings and construct an array on NEW cigar elements
    #

    LOG.debug("CIGAR CONVERSION : PHASE 1 : Converting cigar elements")
    new_cigar = _cigar_convert(cigar, chromosome, chain, strand, position)
    LOG.debug("AFTER PHASE 1 : {0} ".format(new_cigar))

    if len(new_cigar) == 1:

        LOG.debug("CIGAR CONVERSION : Skipping to end since only 1 element")

    else:

        #
        # PHASE 2: Remove S if surrounded by M
        #

        LOG.debug("CIGAR CONVERSION : PHASE 2 : Remove S if surrounded by M")
        new_cigar = _cigar_remove_softs_between_m(new_cigar)
        LOG.debug("AFTER PHASE 2 : {0} ".format(new_cigar))

        #
        # PHASE 3: Fix element lengths
        #

        LOG.debug("CIGAR CONVERSION : PHASE 3 : Fix element lengths")
        new_cigar = _cigar_fix_lengths(new_cigar, sequence)
        LOG.debug("AFTER PHASE 3 : {0} ".format(new_cigar))

        #
        # PHASE 4: Combine consecutive matching elements
        #

        LOG.debug("CIGAR CONVERSION : PHASE 4 : Combining elements")
        new_cigar = _cigar_combine_consecutive(new_cigar)
        LOG.debug("AFTER PHASE 4 : {0} ".format(new_cigar))

        #
        # PHASE 5: Combine consecutive matching elements
        #

        LOG.debug("CIGAR CONVERSION : PHASE 5 : Fix pre and post Ms")
        new_cigar = _cigar_fix_pre_and_post_M(new_cigar)
        LOG.debug("AFTER PHASE 5 : {0} ".format(new_cigar))

    #
    # Final pass through CIGAR string
    #
    # test cigar string length
    #
    # SEQ: segment SEQuence. This field can be a '*' when the sequence is not stored. If not a '*',
    # the length of the sequence must equal the sum of lengths of M/I/S/=/X operations in CIGAR.
    # An '=' denotes the base is identical to the reference base. No assumptions can be made on the
    # letter cases.
    #

    LOG.debug("CIGAR CONVERSION : PHASE 6 : Testing length and conversion")

    cigar_seq_length = 0

    # simplify the cigar, throw away the other stuff we used
    simple_cigar = []
    for c in new_cigar:
        simple_cigar.append((CIGAR_C2N[c.code], c.length))
        if c.code in [CIGAR_M, CIGAR_I, CIGAR_S, CIGAR_E, CIGAR_X]:
            cigar_seq_length += c.length

    if cigar_seq_length != len(sequence):
        LOG.debug("CIGAR SEQ LENGTH={0} != SEQ_LEN={1}".format(cigar_seq_length, len(sequence)))
        # not equal according to chain file format, add the clipping length
        simple_cigar.append((CIGAR_s, len(sequence) - cigar_seq_length))

    if old_cigar != cigar_to_string(simple_cigar):
        LOG.debug("old cigar != new cigar")
    else:
        LOG.debug("old cigar == new cigar")

    LOG.debug("CIGAR CONVERSION : {0} ==> {1}".format(old_cigar, cigar_to_string(simple_cigar)))

    LOG.debug(simple_cigar)
    return simple_cigar


if __name__ == '__main__':
    from .g2g_utils import get_logger, configure_logging
    configure_logging(10)
    LOG = get_logger()
    cigarstring = '5I3D4M9D3S104M7D2I'
    cigarlist = _cigar_to_list(cigarstring)
    LOG.debug(cigarstring)
    print cigarlist
    cigar_new = _cigar_remove_softs_between_m(cigarlist)
    #cigar_new = _cigar_fix_pre_and_post_M(cigarlist)
    print cigar_to_string(cigar_new)
    print cigar_new



# -*- coding: utf-8 -*-

#
# Collection of functions related to BED files
#
# 0-based
#

from collections import namedtuple

from .chain import ChainFile
from .exceptions import G2GBedError, G2GLocationError
from .g2g_utils import get_logger
import g2g_fileutils as g2g_fu

LOG = get_logger()

bed_fields = ["chrom", "start", "end", "name", "score", "strand", "extra"]
BEDRecord = namedtuple("BEDRecord", bed_fields)


class BED(object):
    """
    Simple BED object for parsing BED files

    Supports transparent gzip decompression.
    """
    def __init__(self, filename):
        if not filename:
            raise G2GBedError("A filename must be supplied")

        self.filename = filename
        self.current_line = None
        self.current_line_is_bed = False
        self.current_record = None
        self.reader = g2g_fu.open_resource(filename)
        self.nitems = None
        self.current_line_no = 0

    def __iter__(self):
        return self

    def next(self):
        self.current_line = self.reader.next()
        self.current_line_no += 1

        while self.current_line and len(self.current_line.strip()) == 0:
            self.current_line = self.reader.next()
            self.current_line_no += 1

        if self.current_line.startswith("track"):
            self.current_line = self.current_line.strip()
            self.current_line_is_bed = False
            self.current_record = None
            return None

        self.current_line_is_bed = True
        elem = self.current_line.strip().split("\t")

        if not self.nitems:
            self.nitems = len(elem)
        else:
            if self.nitems != len(elem):
               raise G2GBedError("Improperly formatted BED file")

        try:
            bed_data = {'chrom': elem[0],
                        'start': int(elem[1]),
                        'end': int(elem[2]),
                        'name': elem[3] if self.nitems > 3 else None,
                        'score': elem[4] if self.nitems > 4 else None,
                        'strand':  elem[5] if self.nitems > 5 else None,
                        'extra': elem[6:] if self.nitems > 6 else None}

            self.current_record = BEDRecord(**bed_data)
            return self.current_record
        except IndexError, ie:
            LOG.debug(ie.message)
            raise G2GBedError("Improperly formatted BED file, line number: {0}, line: {1}".format(self.current_line_no, self.current_line))
        except ValueError, ve:
            LOG.debug(ve.message)
            raise G2GBedError("Improperly formatted BED file, line number: {0}, line: {1}".format(self.current_line_no, self.current_line))


def convert_bed_file(chain_file, input_file, output_file, reverse=False):
    """
    Convert BED coordinates.

    The mappings of coordinates are stored in the :class:`.chain.ChainFile` object.

    :param chain_file: chain file used for conversion
    :type chain_file: :class:`.chain.ChainFile`
    :param str file_in: the input BED file
    :type file_in: string
    :param file_out: the output BED file
    :type file_out: string
    :param reverse: reverse direction of original chain file
    :type reverse: boolean
    :return: Nothing
    """
    if not isinstance(chain_file, ChainFile):
        chain_file = g2g_fu.check_file(chain_file)

    input_file = g2g_fu.check_file(input_file)
    output_file_name = g2g_fu.check_file(output_file, 'w')
    unmapped_file_name = "{0}.unmapped".format(output_file_name)

    LOG.info("CHAIN FILE: {0}".format(chain_file))
    LOG.info("INPUT FILE: {0}".format(input_file))
    LOG.info("OUTPUT FILE: {0}".format(output_file_name))
    LOG.info("UNMAPPED FILE: {0}".format(unmapped_file_name))

    if not isinstance(chain_file, ChainFile):
        LOG.info("Parsing chain file...")
        chain_file = ChainFile(chain_file, reverse=reverse)
        LOG.info("Chain file parsed")

    bed_out = open(output_file_name, "w")
    bed_unmapped_file = open(unmapped_file_name, "w")

    LOG.info("Converting BED file")

    bed_file = BED(input_file)

    total = 0
    success = 0
    fail = 0

    # BED is 0 based, bx-python is 0 based

    try:
        for record in bed_file:
            # skip over "track" lines
            if not bed_file.current_line_is_bed:
                bed_out.write(bed_file.current_line)
                bed_out.write("\n")
                continue

            total += 1

            mappings = chain_file.find_mappings(record.chrom, record.start, record.end)

            # unmapped
            if mappings:
                success += 1
            else:
                LOG.debug("Fail due to no mappings")
                bed_unmapped_file.write(bed_file.current_line)
                fail += 1
                continue

            start = mappings[0].to_start
            end = mappings[-1].to_end

            LOG.debug("({0}, {1}) => ({2}, {3})".format(record.start, record.end, start, end))

            elems = bed_file.current_line.split()
            elems[1] = start
            elems[2] = end

            bed_out.write("\t".join(map(str, elems)))
            bed_out.write("\n")

        bed_out.close()
        bed_unmapped_file.close()

        LOG.info("Converted {0} of {1} records".format(success, total))
    except G2GLocationError, le:
        LOG.error("{0}: {1}".format(le.message, bed_file.current_line))

    LOG.info('BED file converted')




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_g2gtools
----------------------------------

Tests for `g2gtools` module.
"""

import unittest
import os
import shutil
import pysam
from g2gtools import bamsam, chain


class TestG2gtools(unittest.TestCase):

    def setUp(self):
        self.path = os.path.realpath(__file__)
        self.dir = os.path.dirname(self.path)
        self.dir_parent = os.path.abspath(os.path.join(self.dir, os.pardir))

        self.chain_file = os.path.join(self.dir_parent, "example/example1.chain")
        self.input_file = os.path.join(self.dir_parent, "example/example1.sam")
        self.output_dir = os.path.join(self.dir_parent, "test_example1")

        os.mkdir(self.output_dir)
        self.output_file = os.path.join(self.output_dir, "example1.out.sam")


    def test_example1(self):
        """
        TODO: Not working yet

        :return:
        """
        chain_file = chain.ChainFile(self.chain_file)
        try:
            chain_file.parse()
        except chain.ChainFileException, cfe:
            print cfe

        bamsam.convert_bam_file(chain=chain_file, infile=self.input_file, outfile=self.output_file)

        try:
            sam_file_out = pysam.Samfile(self.output_file, 'rb')
        except:
            sam_file_out = pysam.Samfile(self.output_file, 'r')


        try:
            while 1:
                old_alignment = sam_file_out.next()
                print old_alignment
        except StopIteration:
            print("All reads processed")


        shutil.rmtree(self.output_dir)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import argparse
import sys

from .bed import BED
from .exceptions import G2GError, G2GBAMError, G2GBedError, G2GChainFileError, G2GVCFError, G2GValueError, G2GLocationError, G2GFastaError
from .g2g_utils import Location, app_exit, configure_logging, get_logger, parse_location
from .g2g import file_convert, gtf2chain, db2chain, offset2chain, fasta_extract_exons, fasta_extract_transcripts, dbfetch, fasta_transform
from .gtf_db import gtf2db, get_genes_simple
from .fasta import extract as fasta_extract
from .fasta import extract_id as fasta_extract_id
from .fasta_patch import fasta_patch
from .vcf2chain import vcf2chain
LOG = get_logger()


def command_convert(raw_args, prog=None):
    """
    Convert a file

    Usage: convert [-options] -c <Chain file> -i <input file>

    Required Parameters:
    -c, --chain <Chain file>         input Chain file used in conversion
    -i, --input <input file>         input file to convert
    -o, --output <output file>       file to create

    Optional Parameters:
    -f, --format <bam|sam|gtf|bed>   file format
    -r, --reverse                    reverse the Chain file

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    If no output file is specified [-o], the converted information will be redirected
    to standard out.

    To specify the type of file, use the [-t] option.

    Only the conversion from the same type of file is supported.

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_convert.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-c", "--chain", dest="chain", metavar="Chain_File")
    parser.add_argument("-i", "--input", dest="input", metavar="Input_File")
    parser.add_argument("-o", "--output", dest="output", metavar="Output_File")

    # optional
    parser.add_argument("-f", "--format", dest="format", metavar='bam|sam|gtf|bed', choices=['bam', 'sam', 'gtf', 'bed'])
    parser.add_argument("-r", "--reverse", dest="reverse", action='store_true')

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.chain:
        app_exit("No Chain file was specified.", parser)

    if not args.input:
        app_exit("No input file was specified.", parser)

    if not args.output:
        app_exit("No output file was specified.", parser)

    try:

        file_convert(args.chain, args.input, args.output, args.format, args.reverse)

    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GValueError, ve:
        app_exit(ve, parser)
    except G2GChainFileError, cfe:
        app_exit(cfe, parser)
    except G2GBAMError, e:
        app_exit(e, parser)


def command_gtf2db(raw_args, prog=None):
    """
    Convert a GTF file to a G2G DB file

    Usage: gtf2db [-options] -i <GTF file> -o <G2G DB file>

    Required Parameters:
    -i, --input <GTF file>           GTF file to parse
    -o, --output <G2G DB file>       G2G DB file to create

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_gtf2db.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-i", "--input", dest="input", metavar="GTF_file")
    parser.add_argument("-o", "--output", dest="output", metavar="DB_file")

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.input:
        app_exit("No GTF file was specified.", parser)

    if not args.output:
        app_exit("No output GTG DB file was specified.", parser)

    try:
        gtf2db(args.input, args.output)
    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GValueError, e:
        app_exit(e, parser)
    except G2GChainFileError, e:
        app_exit(e, parser)
    except G2GError, e:
        app_exit(e, parser)


def command_gtf2chain(raw_args, prog=None):
    """
    Convert a GTF file to a new chain file.

    Usage: gtf2chain [-options] -c <Chain file> -i <GTF file>

    Required Parameters:
    -c, --chain <Chain file>         input Chain file used in conversion
    -i, --input <GTF file>           GTF file to parse
    -o, --output <Chain file>        Chain file to create

    Optional Parameters:
    -g, --genes                      Create Gene Chain File, defaults to Transcript chain file

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    """
    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_gtf2chain.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-c", "--chain", dest="chain", metavar="Chain_File")
    parser.add_argument("-i", "--input", dest="input", metavar="Input_File")
    parser.add_argument("-o", "--output", dest="output", metavar="Chain_File")

    # optional
    parser.add_argument("-g", "--genes", dest="genes", action="store_true", default=False)

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.chain:
        app_exit("No from Chain file was specified.", parser)

    if not args.input:
        app_exit("No GTF file was specified.", parser)

    if not args.output:
        app_exit("No output Chain file was specified.", parser)

    try:
        gtf2chain(args.chain, args.input, args.output, args.genes)
    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GValueError, e:
        app_exit(e, parser)
    except G2GChainFileError, e:
        app_exit(e, parser)
    except G2GError, e:
        app_exit(e, parser)


def command_db2chain(raw_args, prog=None):
    """
    Convert a G2G DB file to a Chain file

    Usage: db2chain [-options] -c <Chain file> -i <G2G DB file>

    Required Parameters:
    -c, --chain <Chain file>         input Chain file used in conversion
    -i, --input <G2G DB file>        G2F DB file
    -o, --output <Chain file>        Chain file to create

    Optional Parameters:
    -g, --genes                      Create Gene Chain File, defaults to Transcript chain file

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_db2chain.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-c", "--chain", dest="chain", metavar="Chain_File")
    parser.add_argument("-i", "--input", dest="input", metavar="Input_File")
    parser.add_argument("-o", "--output", dest="output", metavar="Chain_File")

    # optional
    parser.add_argument("-g", "--genes", dest="genes", action="store_true", default=False)

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.chain:
        app_exit("No input Chain file was specified.", parser)

    if not args.input:
        app_exit("No G2G DB file was specified.", parser)

    if not args.output:
        app_exit("No output Chain was specified.", parser)

    try:
        db2chain(args.chain, args.input, args.output, args.genes)
    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GValueError, e:
        app_exit(e, parser)
    except G2GChainFileError, e:
        app_exit(e, parser)


def command_vcf2chain(raw_args, prog=None):
    """
    Convert a VCF file to a Chain file

    Usage: vcf2chain [-options] -i <VCF file> -f <Fasta file> -s <strain>

    Required Parameters:
    -f, --fasta <Fasta file>         Fasta file to parse, BGZIP file supported
    -i, --input <input file>         VCF file to convert
    -o, --output <output file>       Chain file to create
    -s, --strain <strain>            name of strain column in VCF file

    Optional Parameters:
    --diploid                        ignore hetorzygotes, create 2 files
    --keep                           keep track of VCF lines that cannot be converted to Chain file
    --pass                           use only VCF lines that have a PASS for the filter value
    --quality                        filter on quality, FI=PASS

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_vcf2chain.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-f", "--fasta", dest="fasta", metavar="Fasta_File")
    parser.add_argument("-i", "--input", dest="input", metavar="Input_File")
    parser.add_argument("-o", "--output", dest="output", metavar="Chain_File")
    parser.add_argument("-s", "--strain", dest="strain", metavar="strain")

    # optional
    parser.add_argument("--diploid", dest="diploid", action='store_true')
    parser.add_argument("--keep", dest="keep", action='store_true')
    parser.add_argument("--pass", dest="passed", action='store_true')
    parser.add_argument("--quality", dest="quality", action='store_true')

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.input:
        app_exit("No VCF file was specified.", parser)

    if not args.fasta:
        app_exit("No Fasta file was specified.", parser)

    if not args.output:
        app_exit("No Chain file was specified.", parser)

    if not args.strain:
        app_exit("No strain was specified.", parser)

    try:
        vcf2chain(args.input, args.fasta, args.strain, args.output, args.keep, args.passed, args.quality, args.diploid)
    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GValueError, e:
        app_exit(e, parser)
    except G2GVCFError, e:
        app_exit(e, parser)


def command_offset2chain(raw_args, prog=None):
    """
    Convert Seqnature offset files to a Chain file

    Usage: offset2chain [-options] -f <length_file_of_from-genome> -t <length_file_of_to-genome>

    Required Parameters:
      -f, --from <length_file_of_from-genome>  Chromosome length file of From-genome
      -o, --output <output file>  Chain file to create
      -t, --to   <length_file_of_to-genome>    Chromosome length file of To-genome

    Optional Parameters:
      None

    Help Parameters:
      -h, --help                  Print the help and exit
      -d, --debug                 Turn debugging on, list multiple times for more messages

    Note:
      Consider using the following awk command if you don't have chromosome length files.
      $ awk '/^>/ { if(seqlen) { print seqlen }; { printf "%s\t", substr($1,2) }; seqlen=0; next; } \
        { seqlen=seqlen+length($0) } END { print seqlen }' ${genome_fasta} > ${outfile}

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_offset2chain.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-f", "--from", dest="fromchrom", metavar="Chromosome_File")
    parser.add_argument("-o", "--output", dest="output", metavar="Chain_File")
    parser.add_argument("-t", "--to", dest="tochrom", metavar="Chromosome_File")

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.fromchrom:
        app_exit("No from file was specified.", parser)

    if not args.tochrom:
        app_exit("No from to was specified.", parser)

    if not args.output:
        app_exit("No output file was specified.", parser)

    try:
        offset2chain(args.fromchrom, args.tochrom, args.output)
    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GValueError, e:
        app_exit(e, parser)
    except G2GChainFileError, ce:
        app_exit(ce, parser)


def command_fasta_extract(raw_args, prog=None):
    """
    Extract a sequence from a Fasta file give a location.

    Usage: extract [-options] -i <Fasta file> [-l <location> | -b <BED file> | -db <Database> | -id <id>]

    Required Parameters:
    -i, --input <Fasta file>         Fasta file to extract sequence from, BGZIP files supported
    -l, --location <location>        seqid:start:end
        - OR -
    -b, --bed <BED file>             BED file
        - OR -
    -db, --database <DB file>        Database file
        - OR -
    -id, --identifier <identifier>   Fasta identifier

    Optional Parameters:
    -c, --complement                 return complement of sequence
    -e, --exons                      For use with -db, defaults to transcript extraction
    -g, --genes                      For use with -db, defaults to transcript extraction
    --raw                            show just sequence
    -r, --reverse                    return reverse of sequence
    -rc, --reversecomplement         return reverse complement of sequence
    -t, --transcripts                For use with -db, defaults to transcript extraction

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    Note: Locations specified on command line are 1 based coordinates.
          Locations specified via BED file are 0 based coordinates.

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_fasta_extract.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-i", "--input", dest="fasta", metavar="FASTA_File")
    parser.add_argument("-b", "--bed", dest="bed_file", metavar="BED_File")
    parser.add_argument("-id", "--identifier", dest="id", metavar="identifier")
    parser.add_argument("-l", "--location", dest="location", metavar="chr1:100000-200000")
    parser.add_argument("-db", "--database", dest="database", metavar="Database")

    # optional
    parser.add_argument("-c", "--complement", dest="complement", action="store_true", default=False)
    parser.add_argument("-e", "--exons", dest="exons", action="store_true", default=False)
    parser.add_argument("-g", "--genes", dest="genes", action="store_true", default=False)
    parser.add_argument("-r", "--reverse", dest="reverse", action="store_true", default=False)
    parser.add_argument("--raw", dest="raw", action="store_true", default=False)
    parser.add_argument("-rc", "--reversecomplement", dest="reversecomplement", action="store_true", default=False)
    parser.add_argument("-t", "--transcripts", dest="transcripts", action="store_true", default=False)

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.fasta:
        app_exit("No Fasta file was specified.", parser)

    flag_count = 0
    flag_count += (1 if args.complement else 0)
    flag_count += (1 if args.reverse else 0)
    flag_count += (1 if args.reversecomplement else 0)

    if flag_count > 1:
        app_exit("Please specify only one of: -c, -r, -rc", parser)

    reverse = (args.reverse or args.reversecomplement)
    complement = (args.complement or args.reversecomplement)

    if args.database:
        # allow only either exon, gene, or transcript extraction, bot all
        output_count = 0
        output_count += (1 if args.genes else 0)
        output_count += (1 if args.transcripts else 0)
        output_count += (1 if args.exons else 0)

        if output_count == 0:
            app_exit("Please specify one of: -e, -g, or -t", parser)
        elif output_count > 1:
            app_exit("Please specify only one of: -e, -g, or -t", parser)

    location_count = 0
    location_count += (1 if args.location else 0)
    location_count += (1 if args.bed_file else 0)
    location_count += (1 if args.database else 0)
    location_count += (1 if args.id else 0)

    if location_count > 1:
        app_exit("Please specify only one of: -l, -b, -id, or -db", parser)

    all_locations = []

    try:
        if args.location:
            all_locations.append(parse_location(args.location, 1))

        elif args.bed_file:
            if flag_count == 1:
                app_exit("Options -c, -r, -rc cannot be used with BED file.", parser)

            bed_file = BED(args.bed_file)
            for bed_rec in bed_file:
                if bed_file.current_line_is_bed:
                    strand = bed_rec.strand if bed_rec.strand else '+'
                    all_locations.append(Location(bed_rec.chrom, bed_rec.start, bed_rec.end, strand, bed_rec.name, 0))

        elif args.database:
            if flag_count == 1:
                app_exit("Options -c, -r, -rc cannot be used with Database.", parser)

            if args.genes:
                LOG.info("Extracting genes from database...")
                genes = get_genes_simple(args.database)

                LOG.debug("Genes extracted, transforming to locations")
                for gene in genes:
                    l = Location(gene.seqid, gene.start-1, gene.end, gene.strand, name=gene.ensembl_id, original_base=1)
                    all_locations.append(l)

                LOG.info("Extracting genes sequences...")
                fasta_extract(args.fasta, all_locations, None, reverse=reverse, complement=complement, raw=args.raw)
                LOG.info("Extracting genes done")

            elif args.transcripts:
                fasta_extract_transcripts(args.fasta, args.database, None, raw=args.raw)

            elif args.exons:
                fasta_extract_exons(args.fasta, args.database, None, raw=args.raw)
            else:
                LOG.error("Nothing to do")

            return

        elif args.id:
            fasta_extract_id(args.fasta, args.id, output=None, reverse=reverse, complement=complement, raw=args.raw)
            return

        fasta_extract(args.fasta, all_locations, output=None, reverse=reverse, complement=complement, raw=args.raw)
    except KeyboardInterrupt, ki:
        LOG.info("Handled Keyboard Interrupt")
    except G2GValueError, ve:
        app_exit(ve.msg)
    except G2GBedError, be:
        app_exit(be.msg)
    except G2GLocationError, le:
        app_exit(le.msg)
    except G2GFastaError, fe:
        app_exit(fe.msg)


def command_fasta_transform(raw_args, prog=None):
    """
    Extract a sequence from a Fasta file give a location, and replace the from
    newly formatted chain file.

    Usage: transform [-options] -c <Chain file> -i <Fasta file> [-l <location> | -b <BED file>]

    Required Parameters:
    -c, --chain <Chain file>         input Chain file
    -i, --input <Fasta file>         Fasta file to extract sequence from
    -l, --location <location>        seqid:start:end
        - OR -
    -b, --bed <BED file>             BED file
    -o, --output <Output file>       file to output to

    Optional Parameters:
    --bgzip                          compress and index output
    -c, --complement                 return complement of sequence
    -r, --reverse                    reverse the chain file

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    Note: --bgzip can be potentially slow depending on the size of the file

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_fasta_transform.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-c", "--chain", dest="chain", metavar="Chain_File")
    parser.add_argument("-i", "--input", dest="fasta", metavar="FASTA_File")
    parser.add_argument("-b", "--bed", dest="bed_file", metavar="BED_File")
    parser.add_argument("-l", "--location", dest="location", metavar="chr1:100000-200000")
    parser.add_argument("-o", "--output", dest="output", metavar="Output_File")

    # optional
    parser.add_argument("--bgzip", dest="bgzip", action='store_true')
    parser.add_argument("--diploid", dest="diploid", action='store_true')
    parser.add_argument("-r", "--reverse", dest="reverse", action="store_true", default=False)

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.chain:
        app_exit("No Chain file was specified.", parser)

    if not args.fasta:
        app_exit("No input Fasta file was specified.", parser)

    if not args.output:
        app_exit("No output Fasta file was specified.", parser)

    try:
        all_locations = None

        if args.location and args.bed_file:
            app_exit("Please use either a location or a BED file.", parser)
        elif args.location and not args.bed_file:
            all_locations = [parse_location(args.location)]
        elif not args.location and args.bed_file:
            bed_file = BED(args.bed_file)
            all_locations = []
            for bed_rec in bed_file:
                if bed_file.current_line_is_bed:
                    strand = bed_rec.strand if bed_rec.strand else '+'
                    all_locations.append(Location(bed_rec.chrom, bed_rec.start, bed_rec.end, strand))

        fasta_transform(args.fasta, args.chain, all_locations, args.output, args.bgzip, args.reverse)
    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GChainFileError, le:
        app_exit("Chain File error")
    except G2GValueError, le:
        app_exit(le.msg)
    except G2GLocationError, le:
        app_exit("Location error")
    except G2GFastaError, fe:
        app_exit("Fasta error")


def command_fasta_patch(raw_args, prog=None):
    """
    Convert a Fasta file into a patched (SNP'd) Fasta file

    Usage: patch [-options] -i <Fasta file> -v <VCF file> -s <strain> -o <Fasta File>

    Required Parameters:
    -i, --input <Fasta file>         Fasta file to patch
    -o, --output <Output file>       Fasta file to output to
    -s, --strain <strain>            name of strain column in VCF file
    -v, --vcf <input file>           VCF file to parse SNPs

    Optional Parameters:
    --bgzip                          compress and index output
    --diploid                        ignore hetorzygotes, create 2 files
    -n, --num_processes <number>     the number of processes to use, defaults to the number of cores
    -p, --pass                       use only VCF lines that have a PASS for the filter value
    --quality                        filter on quality, FI=PASS

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    Note: --bgzip can be potentially slow depending on the size of the file

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_fasta_patch.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-i", "--input", dest="fasta", metavar="FASTA_File")
    parser.add_argument("-o", "--output", dest="output", metavar="Output_File")
    parser.add_argument("-s", "--strain", dest="strain", metavar="strain")
    parser.add_argument("-v", "--vcf", dest="vcf", metavar="Input_File")

    # optional
    parser.add_argument("--bgzip", dest="bgzip", action='store_true')
    parser.add_argument("--diploid", dest="diploid", action='store_true')
    parser.add_argument("-n", "--numprocesses", type=int, dest="numprocesses", metavar="number_of_processes")
    parser.add_argument("-p", "--pass", dest="passonly", action='store_true')
    parser.add_argument("--quality", dest="quality", action='store_true')

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.fasta:
        app_exit("No input Fasta file was specified.", parser)

    if not args.vcf:
        app_exit("No VCF file was specified.", parser)

    if not args.strain:
        app_exit("No strain was specified.", parser)

    if not args.output:
        app_exit("No output Fasta file was specified.", parser)

    try:
        fasta_patch(args.fasta, args.vcf, args.strain, args.output, args.bgzip, args.numprocesses, args.passonly, args.quality, args.diploid)
    except KeyboardInterrupt, ki:
        app_exit(ki, parser)
    except G2GValueError, e:
        app_exit(e, parser)
    except G2GVCFError, e:
        app_exit(e, parser)
    except G2GError, e:
        sys.exit(1)


def command_dbfetch(raw_args, prog=None):
    """
    Retrieve information from a G2G Database

    Usage: dbfetch [-options] -i <G2G DB file> [-l <location> | -b <BED file>]

    Required Parameters:
    -i, --input <G2G DB file>        G2F DB file
    -l, --location <location>        seqid:start:end
        - OR -
    -b, --bed <BED file>             BED file
        - OR -
    --id <Ensembl ID>                Ensembl ID


    Optional Parameters:
    -c, --inclusive                  inclusive or exclusive
    -e, --exons                      output exon information
    -g, --genes                      output gene information
    -o, --output <Output file>       redirect output to a file
    -t, --transcripts                output transcript information

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_dbfetch.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-i", "--input", dest="input", metavar="Input_File")
    parser.add_argument("-b", "--bed", dest="bed_file", metavar="BED_File")
    parser.add_argument("-l", "--location", dest="location", metavar="chr1:100000-200000")
    parser.add_argument("--id", dest="id", metavar="ENSMUSG00000000000")

    # optional
    parser.add_argument("-c", "--inclusive", dest="inclusive", action="store_true", default=False)
    parser.add_argument("-e", "--exons", dest="exons", action="store_true", default=False)
    parser.add_argument("-g", "--genes", dest="genes", action="store_true", default=False)
    parser.add_argument("-o", "--output", dest="output", metavar="Output_File")
    parser.add_argument("-t", "--transcripts", dest="transcripts", action="store_true", default=False)

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    if not args.input:
        app_exit("No G2G DB file was specified.", parser)

    location_count = 0
    location_count += (1 if args.location else 0)
    location_count += (1 if args.bed_file else 0)
    location_count += (1 if args.id else 0)

    if location_count != 1:
        app_exit("Please specify either -l or -b or --id", parser)

    if (args.genes and args.transcripts and args.exons) or (not args.genes and not args.transcripts and not args.exons):
        args.genes = True
        args.transcripts = True
        args.exons = True

    all_locations = []

    try:
        if args.id:
            dbfetch(args.input, args.output, ids=[args.id], display_genes=args.genes, display_transcripts=args.transcripts, display_exons=args.exons)
        else:
            if args.location:
                all_locations.append(parse_location(args.location, 1))

            elif args.bed_file:
                bed_file = BED(args.bed_file)
                for bed_rec in bed_file:
                    if bed_file.current_line_is_bed:
                        strand = bed_rec.strand if bed_rec.strand else '+'
                        all_locations.append(Location(bed_rec.chrom, bed_rec.start, bed_rec.end, strand, 0))

            dbfetch(args.input, args.output, locations=all_locations, display_genes=args.genes, display_transcripts=args.transcripts, display_exons=args.exons)
    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GValueError, ve:
        app_exit(ve.msg)
    except G2GBedError, be:
        app_exit(be.msg)
    except G2GLocationError, le:
        app_exit(le.msg)
    except G2GFastaError, fe:
        app_exit(fe.msg)


def command_parse_location(raw_args, prog=None):
    """
    Parse a location from the command line.  Useful for verifying locations.

    Usage: parse [-options] -l <location>

    Required Parameters:
    -l, --location <location>        seqid:start:end

    Optional Parameters:
    None

    Help Parameters:
    -h, --help                       print the help and exit
    -d, --debug                      turn debugging on, list multiple times for more messages

    """

    if prog:
        parser = argparse.ArgumentParser(prog=prog, add_help=False)
    else:
        parser = argparse.ArgumentParser(add_help=False)

    def print_message(message):
        print message
        print command_fasta_extract.__doc__
        sys.exit()

    parser.error = print_message

    # required
    parser.add_argument("-l", "--location", dest="location", metavar="chr1:100000-200000")

    # optional

    # debugging and help
    parser.add_argument("-h", "--help", dest="help", action='store_true')
    parser.add_argument("-d", "--debug", dest="debug", action="count", default=0)

    args = parser.parse_args(raw_args)

    configure_logging(args.debug)

    if args.help:
        app_exit("", parser)

    try:
        if args.location:
            LOG.debug("Input: {0}".format(args.location))
            loc = parse_location(args.location, 1)
            LOG.debug(loc.__repr__())
            print str(loc)

    except KeyboardInterrupt, ki:
        LOG.debug(ki)
    except G2GValueError, ve:
        app_exit(ve.msg)
    except G2GBedError, be:
        app_exit(be.msg)
    except G2GLocationError, le:
        app_exit(le.msg)
    except G2GFastaError, fe:
        app_exit(fe.msg)



# -*- coding: utf-8 -*-

#
# Collection of functions related to Sequences
#

import re
from string import maketrans

BASES = re.compile(r"([ACTGNactgn]+)")
TRANS = maketrans('ACTGNactgn', 'TGACNtgacn')


def reverse_sequence(sequence):
    """
    Return the reverse of sequence

    :param sequence: either a string or Sequence object
    :return: the reverse string or Sequence object
    """
    if isinstance(sequence, Sequence):
        return Sequence(seq=sequence.seq[::-1], id=sequence.id, start=sequence.start, end=sequence.end, strand=sequence.strand)
    elif isinstance(sequence, str):
        return sequence[::-1]
    else:
        raise ValueError("Cannot complement object of {0}, expecting string or Sequence object".format(type(sequence)))


def complement_sequence(sequence):
    """
    Return the complement of sequence

    :param sequence: either a string or Sequence object
    :return: the complement string or Sequence object
    """
    val = str(sequence)

    if len(re.findall(BASES, val)) == 1:  # finds invalid characters if > 1
        val = str(sequence).translate(TRANS)
    else:
        matches = re.findall(BASES, val)
        position = len(matches[0])
        raise ValueError("Sequence contains non-DNA character '{0}' at position {1:n}\n".format(val[position], position + 1))

    if isinstance(sequence, Sequence):
        return Sequence(seq=val, id=sequence.id, start=sequence.start, end=sequence.end, strand=sequence.strand)
    elif isinstance(sequence, str):
        return val
    else:
        raise ValueError("Cannot complement object of {0}, expecting string or Sequence object".format(type(sequence)))


def reverse_complement_sequence(sequence):
    """
    Return the reverse-complement of sequence

    :param sequence: either a string or Sequence object
    :return: the reverse-complement string or Sequence object
    """
    return reverse_sequence(complement_sequence(sequence))


class Sequence(object):
    """
    Simple class to abstract sequence information.

    Simplest form is:
        seq = Sequence("ACGTACGT")

    Can also take the following keyword arguments:

    id = identifier
    seq = string of ACGT
    start, end = positions
    strand
    """
    def __init__(self, *args, **kwargs):
        self.seq = None
        self.id = None
        self.start = None
        self.end = None
        self.strand = None

        #args -- tuple of anonymous arguments
        #kwargs -- dictionary of named arguments

        if 'seq' in kwargs and len(args) == 1:
            raise ValueError("Incorrect values passed to {0}".format(self.__class__))

        if len(args) == 1:
            self.seq = args[0]

        if 'seq' in kwargs:
            self.seq = kwargs['seq']
        if 'id' in kwargs:
            self.id = kwargs['id']
        if 'start' in kwargs:
            self.start = kwargs['start']
        if 'end' in kwargs:
            self.end = kwargs['end']
        if 'strand' in kwargs:
            self.strand = kwargs['strand']

    def __getitem__(self, n):
        if isinstance(n, slice):
            start, stop, step = n.indices(len(self))
            if stop == -1:
                stop = start
            else:
                stop = len(self) - stop
            if self.start and self.end:
                return self.__class__(seq=self.seq[n.start:n.stop:n.step], id=self.id,
                                  start=self.start + start, end=self.end - stop, strand=self.strand)
            else:
                return self.__class__(seq=self.seq[n.start:n.stop:n.step], id=self.id, strand=self.strand)
        elif isinstance(n, int):
            if n < 0:
                n += len(self)
            if self.start:
                return self.__class__(seq=self.seq[n], id=self.id, start=self.start + n, end=self.start + n, strand=self.strand)
            else:
                return self.__class__(seq=self.seq[n], id=self.id, strand=self.strand)

    def __str__(self):
        return self.seq

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __len__(self):
        return len(self.seq)

    @property
    def info(self):
        """ Return the fancy name for the sequence, including start, end
        """
        if len(self) > 50:
            # Shows the last three letters as it is often useful to see if there
            # is a stop codon at the end of a sequence.
            # Note total length is 44+3+3=50
            return "%s(id='%s', '%s...%s')" % (self.__class__.__name__, self.id, str(self)[:44], str(self)[-3:])

        return "%s(id='%s', '%s')" % (self.__class__.__name__, self.id, repr(self.seq))





# -*- coding: utf-8 -*-

from collections import OrderedDict
import os
import sys
import time

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from .bamsam import convert_bam_file
from .bed import convert_bed_file
from .chain import CHAIN_STRING, ChainFile, ChainIter, ChainEntry, collapse_entries
from .exceptions import G2GChainFileError, G2GVCFError, G2GValueError, G2GLocationError, G2GFastaError, G2GError
from .gtf import convert_gtf_file
from .gtf_db import get_genes_simple, get_transcripts_simple, gtf2db, get_genes, get_gene
from .g2g_utils import format_time, get_logger, parse_location, Location, merge_dicts, reverse_complement_sequence, wrap_sequence
import g2g_fileutils as g2g_fu
from .seq_offsets import offset_parse_chromosomes
from .vcf import VCF, parse_gt_new

from pysam import FastaFile
from pysam import TabixFile
import pysam

LOG = get_logger()


def file_convert(chain_file, input_file, output_file=None, file_format=None, reverse=False):
    """
    Convert a file

    :param chain_file:
    :param input_file:
    :param output_file:
    :param file_format:
    :return:
    """
    start = time.time()

    if file_format:
        file_format = file_format.upper()
        if file_format not in ['BED', 'BAM', 'SAM', 'GTF']:
            raise G2GValueError("Only BAM/SAM to BAM/SAM, GTF to GTF, or BED to BED are supported")
    else:
        # try to determine the type from the input
        file_all_caps = input_file.upper()
        if file_all_caps.endswith(('BAM', 'SAM')):
            file_format = 'BAM'
        elif file_all_caps.endswith('BED'):
            file_format = 'BED'
        elif file_all_caps.endswith('GTF'):
            file_format = 'GTF'
        else:
            raise G2GValueError("File format cannot be determined, please specify.")

    if file_format in ['BAM', 'SAM']:
        convert_bam_file(chain_file=chain_file, file_in=input_file, file_out=output_file, reverse=reverse)
    elif file_format in ['GTF']:
        convert_gtf_file(chain_file=chain_file, input_file=input_file, output_file=output_file, reverse=reverse)
    elif file_format in ['BED']:
        convert_bed_file(chain_file=chain_file, input_file=input_file, output_file=output_file, reverse=reverse)
    else:
        raise G2GValueError("Only BAM/SAM to BAM/SAM, GTF to GTF, or BED to BED are supported")

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))


def gtf2chain(chain_file, input_file, output_file, chain_genes=False):
    """

    :param chain_file:
    :param input_file:
    :param output_file:
    :param chain_genes:
    :return:
    """
    start = time.time()
    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))

    chain_file = g2g_fu.check_file(chain_file)
    input_file = g2g_fu.check_file(input_file)
    output_file = g2g_fu.check_file(output_file, 'w')
    output_file_dir = os.path.dirname(output_file)

    LOG.info("GTF FILE: {0}".format(input_file))
    LOG.info("FROM CHAIN FILE: {0}".format(chain_file))
    LOG.info("TO CHAIN FILE: {0}".format(output_file))

    temp_db = g2g_fu.gen_file_name("_g2gtempfile", output_file_dir, ".db3")

    gtf2db(input_file, temp_db)

    db2chain(chain_file, temp_db, output_file, chain_genes)

    g2g_fu.delete_file(temp_db)

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))


def convert_gene_transcripts(gene, chain_file, output_file):
    """
    Convert a gene's transcripts from the GTG DB file and chain file.

    :param gene: the gene information
    :param chain_file: the chain file
    :param output_file: the new chain file
    :return:
    """
    LOG.debug("Gene = {0}".format(str(gene)))

    if not isinstance(chain_file, ChainFile):
        chain_file = ChainFile(chain_file)

    transcripts = gene['transcripts']
    for t in transcripts:
        transcript = transcripts[t]
        LOG.debug("Transcript = {0}".format(str(transcript)))

        chain_entries = []
        from_start = None
        to_start = None

        for exon_id in transcript['exons']:
            exon = transcript['exons'][exon_id]
            LOG.debug("Exon = {0}".format(str(exon)))

            mappings = chain_file.find_mappings(gene['chrom'], exon['start'], exon['end'])

            if mappings and len(mappings) > 0:
                if not from_start:
                    from_start = mappings[0].from_start
                    to_start = mappings[0].to_start

                if len(mappings) == 1:
                    m = mappings[0]
                    c = ChainEntry()
                    #c.from_start = m.from_start
                    #c.from_end = m.from_end
                    #c.to_start = m.to_start
                    #c.to_end = m.to_end
                    c.lines.append([m.from_end - m.from_start])
                    chain_entries.append(c)
                else:
                    c = ChainEntry()
                    #c.from_start = mappings[0].from_start
                    #c.from_end = mappings[-1].from_end
                    #c.to_start = mappings[0].to_start
                    #c.to_end = mappings[-1].to_end

                    prev_mapping = None
                    sum_size = 0
                    sum_dq = 0
                    sum_dt = 0
                    dq = 0
                    prev_dq = 0
                    dt = 0
                    prev_dt = 0

                    for m in mappings:
                        if not prev_mapping:
                            prev_mapping = m
                        else:
                            prev_dt = dt
                            prev_dq = dq
                            chain_size = prev_mapping.from_end - prev_mapping.from_start
                            dt = m.from_start - prev_mapping.from_end
                            dq = m.to_start - prev_mapping.to_end

                            if dt > 0:
                                chain_size += prev_dq

                            sum_size += chain_size
                            sum_dq += dq
                            sum_dt += dt

                            c.lines.append([chain_size, dt, dq])
                            LOG.debug(c.lines[-1])
                            prev_mapping = m

                    chain_size = mappings[-1].from_end - mappings[-1].from_start
                    if dt > 0:
                        chain_size += dq
                        sum_size += dq

                    c.lines.append([chain_size])
                    chain_entries.append(c)

        # collapse exons
        if chain_entries and len(chain_entries) > 0:
            chain_entries = collapse_entries(chain_entries)
            sum_size = 0
            sum_dq = 0
            sum_dt = 0
            lines = []

            for line in chain_entries[0].lines:
                sum_size += line[0]
                if len(line) > 1:
                    sum_dq += line[1]
                    sum_dt += line[2]
                lines.append('\t'.join(map(str, line)))

            if output_file:
                outf = open(output_file, "a")
            else:
                outf = sys.stdout

            outf.write(CHAIN_STRING.format(CHAIN_STRING,
                        from_chr=t, from_length=sum_size + sum_dq,
                        from_start=0, from_end=sum_size + sum_dq,
                        to_chr=t, to_length=sum_size + sum_dt,
                        to_start=0, to_end=sum_size + sum_dt, id=t))
            outf.write("\n")
            outf.write("\n".join(lines))
            outf.write("\n")

            outf.close()


def db2chain(chain_file, input_file, output_file, chain_genes=False):
    """

    :param chain_file:
    :param input_file:
    :param output_file:
    :param chain_genes:
    :return:
    """
    start = time.time()

    if not isinstance(chain_file, ChainFile):
        chain_file = g2g_fu.check_file(chain_file)

    input_file = g2g_fu.check_file(input_file)
    output_file_name = g2g_fu.check_file(output_file, 'w')

    LOG.info("CHAIN FILE: {0}".format(chain_file))
    LOG.info("INPUT FILE: {0}".format(input_file))
    LOG.info("OUTPUT FILE: {0}".format(output_file_name))

    if chain_genes:
        LOG.info("CHAIN TYPE: GENES")
    else:
        LOG.info("CHAIN TYPE: TRANSCRIPTS")

    if not isinstance(chain_file, ChainFile):
        LOG.info("Parsing chain file...")
        chain_file = ChainFile(chain_file)
        LOG.info("Chain file parsed")

    LOG.info('Creating new chain file...')

    if chain_genes:
        LOG.debug("Generating chain for genes")

        for chromosome in chain_file.get_seqids():
            LOG.debug("Generating chain for genes in chromosome {0}".format(chromosome))

            for i, gene in enumerate(get_genes_simple(input_file, location=Location(chromosome))):
                LOG.debug("\n{0}".format(gene))
                chain_entries = []
                from_start = None
                to_start = None
                from_end = None
                to_end = None

                mappings = chain_file.find_mappings(gene.seqid, gene.start, gene.end)

                if gene.strand == 1:
                    if mappings and len(mappings) > 0:
                        if not from_start:
                            from_start = mappings[0].from_start
                            to_start = mappings[0].to_start

                        if len(mappings) == 1:
                            m = mappings[0]
                            c = ChainEntry()
                            c.lines.append([m.from_end - m.from_start])
                            chain_entries.append(c)
                        else:
                            c = ChainEntry()

                            prev_mapping = None
                            sum_size = 0
                            sum_dq = 0
                            sum_dt = 0
                            dq = 0
                            prev_dq = 0
                            dt = 0
                            prev_dt = 0

                            for m in mappings:
                                if not prev_mapping:
                                    prev_mapping = m
                                else:
                                    prev_dt = dt
                                    prev_dq = dq
                                    chain_size = prev_mapping.from_end - prev_mapping.from_start
                                    dt = m.from_start - prev_mapping.from_end
                                    dq = m.to_start - prev_mapping.to_end

                                    if dt > 0:
                                        chain_size += prev_dq

                                    sum_size += chain_size
                                    sum_dq += dq
                                    sum_dt += dt

                                    c.lines.append([chain_size, dt, dq])
                                    LOG.debug(c.lines[-1])
                                    prev_mapping = m

                            chain_size = mappings[-1].from_end - mappings[-1].from_start
                            if dt > 0:
                                chain_size += dq
                                sum_size += dq

                            c.lines.append([chain_size])
                            chain_entries.append(c)
                else:
                    if mappings and len(mappings) > 0:
                        if not from_end:
                            from_end = mappings[-1].from_end
                            to_end = mappings[-1].to_end

                        if len(mappings) == 1:
                            m = mappings[0]
                            c = ChainEntry()
                            c.lines.append([m.from_end - m.from_start])
                            chain_entries.append(c)
                        else:
                            c = ChainEntry()

                            prev_mapping = None
                            sum_size = 0
                            sum_dq = 0
                            sum_dt = 0
                            dq = 0
                            prev_dq = 0
                            dt = 0
                            prev_dt = 0

                            # reverse
                            mappings = mappings[::-1]

                            for m in mappings:
                                LOG.debug("CURRENT MAPPING: {0}".format(m))
                                if not prev_mapping:
                                    prev_mapping = m
                                else:
                                    LOG.debug("PREV MAPPING: {0}".format(prev_mapping))
                                    prev_dt = dt
                                    prev_dq = dq
                                    chain_size = prev_mapping.from_end - prev_mapping.from_start
                                    #dt = m.from_start - prev_mapping.from_end
                                    #dq = m.to_start - prev_mapping.to_end
                                    dt = prev_mapping.from_start - m.from_end
                                    dq = prev_mapping.to_start - m.to_end
                                    LOG.debug("dt={0}, dq={1}".format(dt, dq))

                               #if dt > 0:
                                    #    LOG.debug("DT > 0, ADDING to current chain_size {0}".format(chain_size))
                                    #    chain_size += prev_dq

                                    sum_size += chain_size
                                    sum_dq += dq
                                    sum_dt += dt

                                    c.lines.append([chain_size, dt, dq])
                                    LOG.debug(c.lines[-1])
                                    prev_mapping = m

                            LOG.debug("finding last...{0}".format(mappings[-1]))
                            chain_size = mappings[-1].from_end - mappings[-1].from_start
                            #if dt > 0:
                            #    LOG.debug("WHOA {0}".format(dt))
                            #    LOG.debug("DT > 0, ADDING to current chain_size {0}".format(chain_size))
                            #    chain_size += dq
                            #    sum_size += dq

                            c.lines.append([chain_size])
                            LOG.debug(c.lines[-1])
                            chain_entries.append(c)

                if chain_entries and len(chain_entries) > 0:
                    sum_size = 0
                    sum_dq = 0
                    sum_dt = 0
                    lines = []

                    for line in chain_entries[0].lines:
                        sum_size += line[0]
                        if len(line) > 1:
                            sum_dq += line[1]
                            sum_dt += line[2]
                        lines.append('\t'.join(map(str, line)))

                    if output_file:
                        outf = open(output_file, "a")
                    else:
                        outf = sys.stdout

                outf.write(CHAIN_STRING.format(CHAIN_STRING,
                            from_chr=gene.seqid, from_length=sum_size + sum_dq,
                            from_start=0, from_end=sum_size + sum_dq,
                            to_chr=gene.seqid, to_length=sum_size + sum_dt,
                            to_start=0, to_end=sum_size + sum_dt, id=gene.ensembl_id))
                outf.write("\n")
                outf.write("\n".join(lines))
                outf.write("\n")

                outf.close()
    else:
        for chromosome in chain_file.get_seqids():
            LOG.debug("Generating chain for transcripts in chromosome {0}".format(chromosome))

            for i, transcript in enumerate(get_transcripts_simple(input_file, location=Location(chromosome))):
                LOG.debug("Transcript = {0}".format(transcript))
                chain_entries = []
                from_start = None
                to_start = None
                from_end = None
                to_end = None
                transcript.exons = OrderedDict(sorted(transcript.exons.items(), key=lambda x: x[1].exon_number))

                for ensembl_id, exon in transcript.exons.iteritems():
                    LOG.debug("Exon = {0}".format(exon))

                    mappings = chain_file.find_mappings(exon.seqid, exon.start, exon.end)

                    if exon.strand == 1:
                        if mappings and len(mappings) > 0:
                            if not from_start:
                                from_start = mappings[0].from_start
                                to_start = mappings[0].to_start

                            if len(mappings) == 1:
                                m = mappings[0]
                                c = ChainEntry()
                                c.lines.append([m.from_end - m.from_start])
                                chain_entries.append(c)
                            else:
                                c = ChainEntry()

                                prev_mapping = None
                                sum_size = 0
                                sum_dq = 0
                                sum_dt = 0
                                dq = 0
                                prev_dq = 0
                                dt = 0
                                prev_dt = 0

                                for m in mappings:
                                    if not prev_mapping:
                                        prev_mapping = m
                                    else:
                                        prev_dt = dt
                                        prev_dq = dq
                                        chain_size = prev_mapping.from_end - prev_mapping.from_start
                                        dt = m.from_start - prev_mapping.from_end
                                        dq = m.to_start - prev_mapping.to_end

                                        if dt > 0:
                                            chain_size += prev_dq

                                        sum_size += chain_size
                                        sum_dq += dq
                                        sum_dt += dt

                                        c.lines.append([chain_size, dt, dq])
                                        LOG.debug(c.lines[-1])
                                        prev_mapping = m

                                chain_size = mappings[-1].from_end - mappings[-1].from_start
                                if dt > 0:
                                    chain_size += dq
                                    sum_size += dq

                                c.lines.append([chain_size])
                                chain_entries.append(c)
                    else:
                        if mappings and len(mappings) > 0:
                            if not from_end:
                                from_end = mappings[-1].from_end
                                to_end = mappings[-1].to_end

                            if len(mappings) == 1:
                                m = mappings[0]
                                c = ChainEntry()
                                c.lines.append([m.from_end - m.from_start])
                                chain_entries.append(c)
                            else:
                                c = ChainEntry()

                                prev_mapping = None
                                sum_size = 0
                                sum_dq = 0
                                sum_dt = 0
                                dq = 0
                                prev_dq = 0
                                dt = 0
                                prev_dt = 0

                                # reverse
                                mappings = mappings[::-1]

                                for m in mappings:
                                    LOG.debug("CURRENT MAPPING: {0}".format(m))
                                    if not prev_mapping:
                                        prev_mapping = m
                                    else:
                                        LOG.debug("PREV MAPPING: {0}".format(prev_mapping))
                                        prev_dt = dt
                                        prev_dq = dq
                                        chain_size = prev_mapping.from_end - prev_mapping.from_start
                                        #dt = m.from_start - prev_mapping.from_end
                                        #dq = m.to_start - prev_mapping.to_end
                                        dt = prev_mapping.from_start - m.from_end
                                        dq = prev_mapping.to_start - m.to_end
                                        LOG.debug("dt={0}, dq={1}".format(dt, dq))

                                   #if dt > 0:
                                        #    LOG.debug("DT > 0, ADDING to current chain_size {0}".format(chain_size))
                                        #    chain_size += prev_dq

                                        sum_size += chain_size
                                        sum_dq += dq
                                        sum_dt += dt

                                        c.lines.append([chain_size, dt, dq])
                                        LOG.debug(c.lines[-1])
                                        prev_mapping = m

                                LOG.debug("finding last...{0}".format(mappings[-1]))
                                chain_size = mappings[-1].from_end - mappings[-1].from_start
                                #if dt > 0:
                                #    LOG.debug("WHOA {0}".format(dt))
                                #    chain_size += dq
                                #    sum_size += dq

                                c.lines.append([chain_size])
                                LOG.debug(c.lines[-1])
                                chain_entries.append(c)

                # collapse exons
                if chain_entries and len(chain_entries) > 0:
                    LOG.debug('>>>>>>>')
                    for c in chain_entries:
                        LOG.debug(str(c))
                    LOG.debug('>>>>>>>')
                    chain_entries = collapse_entries(chain_entries)
                    sum_size = 0
                    sum_dq = 0
                    sum_dt = 0
                    lines = []

                    for line in chain_entries[0].lines:
                        sum_size += line[0]
                        if len(line) > 1:
                            sum_dq += line[1]
                            sum_dt += line[2]
                        lines.append('\t'.join(map(str, line)))

                    if output_file:
                        outf = open(output_file, "a")
                    else:
                        outf = sys.stdout

                    outf.write(CHAIN_STRING.format(CHAIN_STRING,
                                from_chr=transcript.seqid, from_length=sum_size + sum_dq,
                                from_start=0, from_end=sum_size + sum_dq,
                                to_chr=transcript.seqid, to_length=sum_size + sum_dt,
                                to_start=0, to_end=sum_size + sum_dt, id=transcript.ensembl_id))
                    outf.write("\n")
                    outf.write("\n".join(lines))
                    outf.write("\n")

                    outf.close()

    LOG.info('New chain file created')

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))


def offset_chromosome_to_chain(chromosome, chain_file):
    """
    Convert Seqnature offset file to chain file

    :param chromosome: the chromosome data
    :param chain_file: the chain file to write to
    """
    if not os.path.exists(chromosome['file_path']):
        raise IOError('Cannot find {0}'.format(chromosome['file_path']))

    prev_offset_pos = 0
    prev_offset_amt = 0
    first_offset = 0
    ref_after_pos = 0

    sum_size = 0
    sum_dt = 0
    sum_dq = 0
    chain_data = []

    fd = open(chromosome['file_path'], 'r')

    for line in fd:
        elem = line.strip().split()
        offset_pos = int(elem[0])
        offset_amt = int(elem[1])

        LOG.debug('LINE: {0}'.format(elem))

        chain_size = 0
        dt = 0
        dq = 0

        if prev_offset_pos == 0 and prev_offset_amt == 0:
            # first time through
            LOG.debug("ref_after_pos=" + str(ref_after_pos))
            LOG.debug('offset_pos=' + str(offset_pos) + '\toffset_amt=' + str(offset_amt))
            LOG.debug('prev_pos=' + str(prev_offset_pos) + '\tprev_amt=' + str(prev_offset_amt))
            offset_diff = offset_amt
            LOG.debug('offset_diff= ' + str(offset_diff))

            if offset_amt >= 0:
                dt = offset_amt
                ref_after_pos = offset_pos
            else:
                dq = offset_amt * -1
                ref_after_pos = offset_pos + dq

            chain_size = offset_pos - 1
            first_offset = chain_size
            sum_size = chain_size
            sum_dt = dt
            sum_dq = dq
            chain_data.append([chain_size, dt, dq])

            prev_offset_pos = offset_pos
            prev_offset_amt = offset_amt
            LOG.debug('>> {0}\t{1}\t{2}'.format(sum_size, sum_dt, sum_dq))
        else:
            offset_diff = offset_amt - prev_offset_amt

            LOG.debug("PREV ref_after_pos=" + str(ref_after_pos))
            LOG.debug("offset_pos=" + str(offset_pos))

            if offset_diff >= 0:
                dt = offset_diff
                chain_size = offset_pos - ref_after_pos
                ref_after_pos = offset_pos
            else:
                dq = offset_diff * -1
                chain_size = offset_pos - ref_after_pos
                ref_after_pos = offset_pos + dq

            LOG.debug("ref_after_pos=" + str(ref_after_pos))


            sum_size += chain_size
            sum_dt += dt
            sum_dq += dq

            # fix any 0 length
            if chain_size != 0:

                chain_data.append([chain_size, dt, dq])
                LOG.debug(chain_data[-1])
            else:
                LOG.debug("offset_diff=" + str(offset_diff))
                LOG.debug("offset_amt=" + str(offset_amt))
                LOG.debug("prev_offset_amt=" + str(prev_offset_amt))
                LOG.debug("COMBINING!!!!!!!!!!")
                LOG.debug(chain_data[-1])
                LOG.debug("dt={0}".format(dt))
                LOG.debug("dq={0}".format(dq))
                chain_data[-1][1] += dt
                chain_data[-1][2] += dq

            prev_offset_pos = offset_pos
            prev_offset_amt = offset_amt
            LOG.debug('>>SUM {0}\t{1}\t{2}'.format(sum_size, sum_dt, sum_dq))
            LOG.debug('>>    {0}\t{1}\t{2}'.format(chain_size, dt, dq))

    # not sure if the following assumption will work in all cases
    # the last line of the chain segment is the last fragment length
    if chromosome['from_length'] > chromosome['to_length']:
        last_chain = chromosome['from_length'] - (sum_size + sum_dt)
        sum_size += last_chain
    else:
        last_chain = chromosome['to_length'] - (sum_size + sum_dq)
        sum_size += last_chain
    chain_fd = sys.stdout
    if chain_file:
        chain_fd = open(chain_file, "a")

    chain_fd.write(CHAIN_STRING.format(CHAIN_STRING,
            from_chr=chromosome['from_chr'], from_length=chromosome['from_length'],
            from_start=0, from_end=sum_size + sum_dt,
            to_chr=chromosome['to_chr'], to_length=chromosome['to_length'],
            to_start=0, to_end=sum_size + sum_dq, id=chromosome['to_chr']))
    chain_fd.write("\n")

    for c in chain_data:
        chain_fd.write("{0}\t{1}\t{2}\n".format(c[0], c[1], c[2]))

    chain_fd.write("{0}\n\n".format(last_chain))

    chain_fd.close()


def offset2chain(from_file, to_file, output_file):
    """
    Convert Seqnature offset files to new chain file.

    :param from_file: from Chromosome File (see docs)
    :param to_file: to Chromosome File (see docs)
    :param output_file: the output chain file
    """
    start = time.time()

    from_file = g2g_fu.check_file(from_file)
    to_file = g2g_fu.check_file(to_file)

    output_file_name = g2g_fu.check_file(output_file, 'w')
    g2g_fu.delete_file(output_file_name)

    LOG.info("FROM FILE: {0}".format(from_file))
    LOG.info("TO FILE: {0}".format(to_file))
    LOG.info("CHAIN FILE: {0}".format(output_file_name))

    LOG.info("Generating chain file...")

    try:
        chromosomes = offset_parse_chromosomes(from_file, to_file)

        for c, chromosome in chromosomes.iteritems():
            LOG.debug('Examining chromosome: {0}'.format(chromosome))
            if chromosome['file_path']:
                offset_chromosome_to_chain(chromosome, output_file)
            else:
                LOG.debug("No file for {0}, so skipping".format(chromosome))

        LOG.info("Chain file created")

    except Exception, e:
        raise G2GChainFileError("Unable to generate chain file")

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))


def fasta_extract_transcripts(fasta_file, database_file, output, raw=False):
    start = time.time()

    if isinstance(fasta_file, FastaFile):
        fasta = fasta_file
    else:
        fasta_file = g2g_fu.check_file(fasta_file)
        fasta = FastaFile(fasta_file)

    database_file = g2g_fu.check_file(database_file)

    fasta_out = sys.stdout

    if output:
        output = g2g_fu.check_file(output, 'w')
        fasta_out = open(output, "w")

    LOG.info("FASTA FILE: {0}".format(fasta.filename))
    LOG.info("DATABASE FILE: {0}".format(database_file))
    LOG.info("OUTPUT FILE: {0}".format(fasta_out.name))

    try:
        transcripts = get_transcripts_simple(database_file)

        for i, transcript in enumerate(transcripts):
            LOG.debug("Transcript={0}".format(transcript))

            if transcript.seqid not in fasta.references:
                continue

            new_sequence = StringIO()
            locations = []
            for ensembl_id, exon in transcript.exons.iteritems():
                LOG.debug("Exon ID={0};{1}".format(ensembl_id, exon))

                partial_seq = fasta.fetch(exon.seqid, exon.start-1, exon.end)
                partial_seq_str = str(partial_seq)

                if transcript.strand == 1:
                    new_sequence.write(partial_seq_str)
                else:
                    partial_seq_str = str(reverse_complement_sequence(partial_seq))
                    new_sequence.write(partial_seq_str)

                LOG.debug("{0}:{1}-{2} (Length: {3})\n{4}".format(exon.seqid, exon.start, exon.end, len(partial_seq), partial_seq_str))
                locations.append("{0}:{1}-{2}".format(exon.seqid, exon.start, exon.end))

            if raw:
                fasta_out.write(new_sequence.getvalue())
            else:
                fasta_id = ">{0} {1}|{2}\n".format(transcript.ensembl_id, '-' if transcript.strand == -1 else '+', "|".join(locations))
                fasta_out.write(fasta_id)

                for line in wrap_sequence(new_sequence.getvalue()):
                    fasta_out.write(line.strip())
                    fasta_out.write('\n')

    except G2GValueError as e:
        LOG.info(e.msg.rstrip())
        raise e
    except G2GFastaError as e:
        LOG.info(e.msg.rstrip())
        raise e

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))


def fasta_extract_exons(fasta_file, database_file, output, raw=False):
    start = time.time()

    if isinstance(fasta_file, FastaFile):
        fasta = fasta_file
    else:
        fasta_file = g2g_fu.check_file(fasta_file)
        fasta = FastaFile(fasta_file)

    database_file = g2g_fu.check_file(database_file)

    fasta_out = sys.stdout

    if output:
        output = g2g_fu.check_file(output, 'w')
        fasta_out = open(output, "w")

    LOG.info("FASTA FILE: {0}".format(fasta.filename))
    LOG.info("DATABASE FILE: {0}".format(database_file))
    LOG.info("OUTPUT FILE: {0}".format(fasta_out.name))

    try:
        transcripts = get_transcripts_simple(database_file)
        for i, transcript in enumerate(transcripts):

            if transcript.seqid not in fasta.references:
                continue

            for ensembl_id, exon in transcript.exons.iteritems():
                LOG.debug("Exon={0}".format(exon))

                partial_seq = fasta.fetch(exon.seqid, exon.start-1, exon.end)
                partial_seq_str = partial_seq

                if transcript.strand == -1:
                    partial_seq_str = str(reverse_complement_sequence(partial_seq))

                LOG.debug("{0}:{1}-{2} (Length: {3})\n{4}".format(exon.seqid, exon.start, exon.end, len(partial_seq), partial_seq_str))

                if raw:
                    fasta_out.write(partial_seq_str)
                else:
                    fasta_id = ">{0} {1}:{2}-{3}\n".format(exon.ensembl_id, exon.seqid, exon.start, exon.end)
                    fasta_out.write(fasta_id)

                    for line in wrap_sequence(partial_seq_str):
                        fasta_out.write(line.strip())
                        fasta_out.write('\n')

    except G2GValueError as e:
        LOG.info(e.msg.rstrip())
        raise e
    except G2GFastaError as e:
        LOG.info(e.msg.rstrip())
        raise e

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))


def dbfetch(input_file, output_file=None, locations=None, ids=None, display_genes=True, display_transcripts=True, display_exons=True, overlap=True):
    """
    """
    start = time.time()

    if input_file and os.path.exists(input_file):
        input_file = os.path.abspath(input_file)
    else:
        raise G2GValueError("The following G2G DB file does not exist: {0}".format(input_file))

    output_file_name = None

    if output_file:
        output_file = os.path.abspath(output_file)
        output_file_dir = os.path.dirname(output_file)
        output_file_name = output_file

        if not os.access(output_file_dir, os.W_OK | os.X_OK):
            raise G2GValueError("Cannot generate file in specified directory: {0}".format(output_file_dir))

    LOG.info("DB FILE: {0}".format(input_file))

    if output_file_name:
        LOG.info("OUTPUT FILE: {0}".format(output_file_name))

    if (locations is None and ids is None) or (locations and ids):
        raise G2GValueError("Can only search by location or ids separately")

    genes = {}
    if locations:
        try:
            for l in locations:
                if not isinstance(l, Location):
                    l = parse_location(l)

                genes = merge_dicts(genes, get_genes(input_file, l))
        except G2GLocationError, le:
            LOG.debug("Unable to parse location, {0}".format(le.message))
            raise le
    else:
        for i in ids:
            genes = merge_dicts(genes, get_gene(input_file, i))

    for k,v in genes.iteritems():
        indent = 0
        if display_genes:
            print v

        for k1, v1 in v.transcripts.iteritems():
            if display_transcripts:
                indent = 1 if display_genes else 0
                print '\t'*indent, str(v1)

            if display_exons:
                if display_genes and display_transcripts:
                    indent = 2
                elif display_genes or display_transcripts:
                    indent = 1
                else:
                    indent = 0

                for k2,v2 in v1.exons.iteritems():
                    print '\t'*indent, str(v2)

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))


def fasta_transform(fasta_file, chain_file, locations, output_file, bgzip=False, reverse=False):
    """

    :param fasta_file:
    :param chain_file:
    :param locations:
    :param output_file:
    :param bgzip:
    :param reverse:
    :return:
    """
    start = time.time()

    if not isinstance(fasta_file, FastaFile):
        fasta_file = g2g_fu.check_file(fasta_file)

    if not isinstance(chain_file, ChainIter):
        chain_file = g2g_fu.check_file(chain_file)

    output_file = g2g_fu.check_file(output_file, 'w')
    g2g_fu.delete_file(output_file)
    g2g_fu.delete_index_files(output_file)

    LOG.info("FASTA FILE: {0}".format(fasta_file))
    LOG.info("CHAIN FILE: {0}".format(chain_file))
    LOG.info("OUTPUT FILE: {0}".format(output_file))
    LOG.info("BGZIP: {0}".format(bgzip))
    LOG.info("REVERSE: {0}".format(reverse))

    if isinstance(fasta_file, FastaFile):
        fasta = fasta_file
    else:
        fasta = FastaFile(fasta_file)

    if not isinstance(chain_file, ChainIter):
        chain_file = ChainIter(chain_file, reverse=reverse)

    seq_ids = []

    if locations:
        LOG.debug("Have locations")
        new_locations = []
        for l in locations:
            if isinstance(l, Location):
                new_locations.append(l)
            else:
                new_locations.append(parse_location(l))
            seq_ids.append(new_locations[-1].seqid)
        locations = new_locations
    else:
        LOG.debug("Calculating locations")
        locations = [parse_location("{0}:1-{1}".format(a, fasta.get_reference_length(a)), 1) for a in fasta.references]
        seq_ids = [a for a in fasta.references]

    temp_output_file = output_file

    if bgzip:
        if g2g_fu.get_extension(output_file) != 'gz':
            output_file = "{0}.gz".format(output_file)
        else:
            temp_output_file = temp_output_file[:-3]

    fasta_out = open(temp_output_file, "w")

    LOG.info("Transforming...")

    chr_info = {}

    try:
        # will need a better way, but for now...
        LOG.info("Parsing chain file...")
        for line in chain_file:
            if len(line) > 7:
                LOG.debug("Adding chromosome {0}".format(chain_file.current_chain_header[1]))
                chr_info[chain_file.current_chain_header[1]] = {'from_size': line[2], 'from_start': line[4], 'from_end': line[5],
                                  'to_size': line[7], 'to_start': line[9], 'to_end': line[10],
                                  'header_chain':chain_file.current_chain_header, 'lines': []}
            else:
                chr_info[chain_file.current_chain_header[1]]['lines'].append(line)

        LOG.info("Chain file parsed")

        insertion_bases = 0
        deletion_bases = 0

        for location in locations:
            LOG.info("Processing chromosome={0}".format(location.seqid))
            LOG.debug("Location: {0}".format(location))

            chrom_size_from = chr_info[location.seqid]['from_size']
            chrom_size_to = chr_info[location.seqid]['to_size']

            last_pos = chr_info[location.seqid]['from_start']
            new_sequence = StringIO()
            chain_file.reset()

            for chain_line in chr_info[location.seqid]['lines']:
                LOG.debug("\nLINE: {0} : {1}".format(chain_file.line_no, chain_line))

                if len(chain_line) == 1:
                    # last line
                    fragment = chain_line[0]

                    partial_seq = fasta.fetch(location.seqid, last_pos, last_pos + fragment)
                    new_sequence.write(str(partial_seq))

                    if len(new_sequence.getvalue()) < chrom_size_to:
                        LOG.warn("Length's do not match, chromosome length in chain: {0}, sequence length: {1}".format(chrom_size_to, len(new_sequence.getvalue())))

                    fasta_out.write(">{0} {1}:{2}-{3}\n".format(location.seqid, location.seqid, chr_info[location.seqid]['from_start'] + 1, chrom_size_to))

                    for l in wrap_sequence(new_sequence.getvalue()):
                        fasta_out.write(l.strip())
                        fasta_out.write('\n')

                    break

                else:

                    # fragment_size dt_size dq_size same_bases dt_bases dq_bases

                    fragment = chain_line[0]
                    dt = chain_line[1 if not reverse else 2]
                    dq = chain_line[2 if not reverse else 1]
                    same = chain_line[3]
                    dt_bases = chain_line[4 if not reverse else 5]
                    dq_bases = chain_line[5 if not reverse else 4]

                    partial_seq = fasta.fetch(location.seqid, last_pos, last_pos + fragment)
                    new_sequence.write(partial_seq)

                    if dq > 0:
                        # insertion
                        LOG.debug("INSERTION")
                        new_sequence.write(dq_bases)
                        LOG.debug("{0}:{1}-{2} (Length: {3})".format(location.seqid, last_pos, last_pos + fragment, len(partial_seq)))
                        if len(partial_seq) > 100:
                            LOG.debug("{0}...{1}".format(partial_seq[:10], partial_seq[-10:]))
                        else:
                            LOG.debug(partial_seq)
                        LOG.debug("Adding {0}".format(dq_bases))
                        LOG.debug("SAME={0}, {1}".format(same, partial_seq[-(len(same)):]))

                        insertion_bases += dq

                    if dt > 0:
                        # deletion
                        LOG.debug("DELETION")
                        last_pos += dt
                        LOG.debug("skipping ahead {0} bases".format(dt))

                        deletion_bases += dt

                    last_pos += fragment

                    LOG.debug("LAST_POS={0}, INSERTIONS={1}, DELETIONS={2}, DIFF={3}".format(last_pos, insertion_bases, deletion_bases, (insertion_bases - deletion_bases)))

        # bgzip and index
        if bgzip:
            LOG.info("Compressing and indexing...")
            g2g_fu.bgzip_index(temp_output_file, output_file, 'fa')

    except G2GLocationError, le:
        LOG.debug("Unable to parse location, {0}".format(le.message))
        raise le
    except G2GValueError as e:
        LOG.debug("Unable to parse alocation, {0}".format(e.message))
        raise e
    except G2GFastaError as e:
        LOG.debug("Unable to parse blocation, {0}".format(e.message))
        raise e
    except TypeError as e:
        LOG.debug("Unable to parse clocation, {0}".format(e.message))
        raise e
        #print str(e)
        #raise G2GError("Improper chain file")

    LOG.info("Transforming complete")

    LOG.info("Execution complete: {0}".format(format_time(start, time.time())))



# -*- coding: utf-8 -*-

# http://genomewiki.ucsc.edu/index.php/Bin_indexing_system

BIN_NEXT_SHIFT = 3    # How much to shift to get to finest bin
BIN_FIRST_SHIFT = 17  # How much to shift to get to next larger bin
#BIN_OFFSETS = [32768+4096+512+64+8+1, 4096+512+64+8+1, 512+64+8+1, 64+8+1, 8+1, 1, 0]
BIN_OFFSETS = [4096+512+64+8+1, 512+64+8+1, 64+8+1, 8+1, 1, 0]
#BIN_OFFSETS = [512+64+8+1, 64+8+1, 8+1, 1, 0]
# for BED (0-based, half-open) or GFF (1-based, closed intervals)
COORD_OFFSETS = {'bed': 0, 'gff': 1}

def bins(start, stop, fmt='gff', one=True):
    """
    Uses the definition of a "genomic bin" described in Fig 7 of
    http://genome.cshlp.org/content/12/6/996.abstract.

    Parameters
    ----------
    one : boolean
        If `one=True` (default), then only return the smallest bin that
        completely contains these coordinates (useful for assigning a single
        bin).

        If `one=False`, then return the set of *all* bins that overlap these
        coordinates (useful for looking for features that could intersect)

    fmt : 'gff' | 'bed'
        This specifies 1-based start coords (gff) or 0-based start coords (bed)
    """
    # Jump to highest resolution bin that will fit these coords (depending on
    # whether we have a BED or GFF-style coordinate).
    start = (start - COORD_OFFSETS[fmt]) >> BIN_FIRST_SHIFT
    stop = (stop) >> BIN_FIRST_SHIFT

    # We always at least fit within the chrom, which is bin 1.
    bins = set([1])

    for offset in BIN_OFFSETS:
        # Since we're going from smallest to largest bins, the first one where
        # the feature's start and stop positions are both within the same bin
        # is the smallest one these coords fit within.
        if one:
            if start == stop:
                # Note that at this point, because of the bit-shifting, `start`
                # is the number of bins (at this current level).  So we need to
                # add it to `offset` to get the actual bin ID.
                return offset + start

        # See the Fig 7 reproduction above to see why range().
        bins.update(list(range(offset + start, offset + stop + 1)))

        # Move to the next level (8x larger bin size; i.e., 2**NEXT_SHIFT
        # larger bin size)
        start >>= BIN_NEXT_SHIFT
        stop >>= BIN_NEXT_SHIFT
    return bins


def bin_info():
    """
    Useful for debugging: how large is each bin, and what are the bin IDs?
    """
    print "level\t#bins\tstart\tend\tsize"
    level = -1
    for i, offset in reversed(list(enumerate(BIN_OFFSETS))):
        level += 1
        number_of_bins = 8 ** level

        binstart = offset
        try:
            binstop = BIN_OFFSETS[i+1]
            binstop = number_of_bins + binstart - 1
        except IndexError:
            binstop = binstart

        bin_size = 2 ** (BIN_FIRST_SHIFT + (i * BIN_NEXT_SHIFT))
        actual_size = bin_size

        # nice formatting
        bin_size, suffix = bin_size / 1024, 'Kb'
        if bin_size >= 1024:
            bin_size, suffix = bin_size / 1024, 'Mb'
        if bin_size >= 1024:
            bin_size, suffix = bin_size / 1024, 'Gb'
        size = '(%s %s)' % (bin_size, suffix)
        actual_size = '%s bp' % (actual_size)

        print('{level}\t{number_of_bins}\t{binstart}\t{binstop}\t{actual_size}\t{size}'.format(**locals()))


if __name__ == "__main__":
    bin_info()




# -*- coding: utf-8 -*-

#
# Collection of g2g utility functions and classes
#

import logging
import os
import re
import sys

from itertools import izip_longest
from string import maketrans

from . import G2GLocationError, G2GValueError

REGEX_LOCATION = re.compile("(\w*)\s*(-|:)?\s*(\d+)\s*(MB|M|K|)?\s*(-|:|)?\s*(\d+|)\s*(MB|M|K|)?", re.IGNORECASE)
REGEX_LOCATION_CHR = re.compile("(CHR|)*\s*([0-9]{1,2}|X|Y|MT)\s*(-|:)?\s*(\d+)\s*(MB|M|K|)?\s*(-|:|)?\s*(\d+|)\s*(MB|M|K|)?", re.IGNORECASE)
BASES = re.compile(r"([ACTGNactgn]+)")
TRANS = maketrans('ACTGNactgn', 'TGACNtgacn')

LOG = logging.getLogger('G2G')


class Location(object):
    def __init__(self, seqid, start=None, end=None, strand='+', name=None, original_base=0):
        if start:
            end = end if end else start + 1
        self.seqid = seqid
        self._strand = None
        self._start = None
        self._end = None
        self.original_base = original_base
        self.strand = strand
        self.start = start
        self.end = end
        self.name = name

    @property
    def strand(self):
        return self._strand

    @strand.setter
    def strand(self, value):
        if value is not None:
            if value in ('+', '1', 1):
                self._strand = '+'
            elif value in ('-', '-1', -1):
                self._strand = '-'
            else:
                raise G2GLocationError("Illegal value for strand {0}, must be +, -, 1, -1".format(value))

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        if value is not None:
            if value < 0:
                raise G2GLocationError("Illegal value for start {0}, start must be >= 0".format(value))
            self._start = value

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        if value is not None:
            if self.start and value <= self.start:
                raise G2GLocationError("Illegal value for end {0}, end must be >= start {1}".format(value, self.start))

            self._end = value
        else:
            if self.start:
                self._end = self.start + 1

    def __str__(self):
        if self.original_base == 1:
            return "{0}:{1}-{2} ({3})".format(self.seqid, self.start+1, self.end, self.strand)
        else:
            return "{0}:{1}-{2} ({3})".format(self.seqid, self.start, self.end, self.strand)

    def __repr__(self):
        return "{0}:{1}-{2} ({3})".format(self.seqid, self.start, self.end, self.strand)


def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


class G2GFormatter(logging.Formatter):

    err_fmt = "ERROR: %(msg)s"
    #dbg_fmt = "DEBUG: %(asctime)s, %(module)s: %(lineno)d: %(msg)s"
    dbg_fmt = "DEBUG: %(asctime)s: %(msg)s"
    info_fmt = "%(msg)s"

    def __init__(self, fmt="%(levelno)s: %(msg)s"):
        logging.Formatter.__init__(self, fmt)

    def format(self, record):

        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.DEBUG:
            self._fmt = G2GFormatter.dbg_fmt

        elif record.levelno == logging.INFO:
            self._fmt = G2GFormatter.info_fmt

        elif record.levelno == logging.ERROR:
            self._fmt = G2GFormatter.err_fmt

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._fmt = format_orig

        return result


def get_logger():
    return LOG


def configure_logging(level):
    """

    :param level: 0=ERROR, 1=INFO, 2+=DEBUG
    """
    global LOG
    LOG = logging.getLogger('G2G')

    handler = logging.StreamHandler()

    mfmt = G2GFormatter()
    handler.setFormatter(mfmt)

    if level == 0:
    #    LOG.setLevel(logging.ERROR)
    #elif level == 1:
        LOG.setLevel(logging.INFO)
    else:
        LOG.setLevel(logging.DEBUG)

    LOG.addHandler(handler)


def app_exit(message='', parser=None):
    """
    Print message, help, and exit.

    :param message: The message to print
    :param parser: the argument parser
    """
    if parser:
        parser.error(message)
    else:
        if message:
            print str(message)

    sys.exit()


def format_time(start, end):
    """
    Format length of time between start and end.

    :param start: the start time
    :param end: the end time
    :return: a formatted string of hours, minutes, and seconds
    """
    hours, rem = divmod(end-start, 3600)
    minutes, seconds = divmod(rem, 60)
    return "{:0>2}:{:0>2}:{:05.2f}".format(int(hours), int(minutes), seconds)


def intersect_regions(chr1, start1, end1, chr2, start2, end2):
    """
    Return the intersection of two regions

    :param chr1: chromosome 1
    :param start1: start position 1
    :param end1: end position 1
    :param chr2: chromosome 2
    :param start2: start position 2
    :param end2: end position 2
    :return: the intersection of the coordinated
    """

    if chr1 != chr2:
        return None

    if int(start1) > int(end2) or int(end1) < int(start2):
        return None

    if int(start1) > int(end1) or int(start2) > int(end2):
        raise G2GLocationError("Start cannot be larger than end")

    return chr1, max(start1, start2), min(end1, end2)


def parse_chromosome(chrom_file):
    '''
    Parse the chromosome for length
    '''
    chromosomes = {}

    try:
        if not os.path.exists(chrom_file):
            raise IOError(chrom_file)

        fd = open(chrom_file, 'r')
        for line in fd:
            elem = line.strip().split()
            try:
                chromosome = elem[0]
                chromosome_length = int(elem[1])
                chromosomes[chromosome] = chromosome_length
            except ValueError, e1:
                pass # most likely header line
        fd.close()
    except IOError, e:
        message = "Error parsing chromosome files: {0}".format(e.message)
        print message
        return {}

    return chromosomes


def get_multiplier(factor):
    """
    Convert the factor into a number.

    :param factor: the string 'mb', 'm', or 'k'
    :return: 10000000, 1000000, 1000 or 1
    """
    if factor.lower() == 'mb':
        return 10000000
    elif factor.lower() == 'm':
        return 1000000
    elif factor.lower() == 'k':
        return 1000

    return 1


def parse_location(location_str, base=0):
    """
    Parse a string and return a location.

    Format expected is seqid|start|end, where | is some sort of delimiter.

    If end is None, 1 based was requested.

    :param location_str: a string representing a genomic location
    :return: a Location
    """
    loc_match = REGEX_LOCATION.match(location_str.strip().replace(',', ''))
    one_base = False

    if loc_match:
        loc_groups = loc_match.groups()
        identifier = loc_groups[0]
        start_base = loc_groups[2]
        start_mult = loc_groups[3]
        end_base = loc_groups[5]
        end_mult = loc_groups[6]

        if not identifier:
            raise G2GLocationError("Cannot parse location")

        if start_base:
            try:
                start_base = int(start_base)
            except ValueError:
                raise G2GLocationError("Start position is not numeric")
        else:
            raise G2GLocationError("Cannot parse start position")

        start = start_base

        if start_mult:
            if start_mult.lower() not in ['mb', 'm', 'k']:
                raise G2GLocationError("Unknown quantifier '{0}".format(start_mult))
            start = start_base * get_multiplier(start_mult)
        else:
            if base == 1:
                start -= 1

        if end_base:
            try:
                end_base = int(end_base)
            except ValueError:
                raise G2GLocationError("End position is not numeric")
        else:
            # special condition to allow for only 1 location to be retrieved
            # or from this start base on

            if not loc_groups[4] and not end_base and not end_mult:
                # only 1 position requested
                one_base = True
            else:
                raise G2GLocationError("Cannot parse end position")

        if end_mult:
            if end_mult.lower() not in ['mb', 'm', 'k']:
                raise G2GLocationError("Unknown quantifier '{0}".format(end_mult))
    else:
        raise G2GLocationError("Cannot parse location '{0}'".format(location_str))

    if one_base:
        end = None
    else:
        end = end_base * get_multiplier(end_mult)

    LOG.debug("identifier={0}, start={1}, end={2}, base={3}".format(identifier, start, end, base))

    return Location(identifier, start, end, '+', original_base=base)


def wrap_sequence(sequence, n=60, fillvalue=''):
    args = [iter(sequence)] * n
    for line in izip_longest(fillvalue=fillvalue, *args):
        yield ''.join(line + ('\n',))


def reverse_sequence(sequence):
    """
    Return the reverse of sequence

    :param sequence: either a string or Sequence object
    :return: the reverse string or Sequence object
    """
    if isinstance(sequence, str):
        return sequence[::-1]
    else:
        raise ValueError("Cannot complement object of {0}, expecting string or Sequence object".format(type(sequence)))


def complement_sequence(sequence):
    """
    Return the complement of sequence

    :param sequence: either a string or Sequence object
    :return: the complement string or Sequence object
    """
    val = str(sequence)

    if len(re.findall(BASES, val)) == 1:  # finds invalid characters if > 1
        val = str(sequence).translate(TRANS)
    else:
        matches = re.findall(BASES, val)
        position = len(matches[0])
        raise ValueError("Sequence contains non-DNA character '{0}' at position {1:n}\n".format(val[position], position + 1))

    if isinstance(sequence, str):
        return val
    else:
        raise ValueError("Cannot complement object of {0}, expecting string or Sequence object".format(type(sequence)))


def reverse_complement_sequence(sequence):
    """
    Return the reverse-complement of sequence

    :param sequence: either a string or Sequence object
    :return: the reverse-complement string or Sequence object
    """
    return reverse_sequence(complement_sequence(sequence))


if __name__ == '__main__':
    l = parse_location("1:0-33")
    print l
    print l.__repr__()

    l = parse_location("1:1-33")
    print l
    print l.__repr__()

    print '1'
    l = Location(1, 22, None, '-')
    print l
    print l.__repr__()

    print 'tryng to set l.start'
    l.start = 2
    print l
    print l.__repr__()
